# Copyright 2017 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for db purge."""

import datetime
import uuid

from oslo_db.sqlalchemy import utils as sqlalchemyutils
from oslo_utils import timeutils
from sqlalchemy.dialects import sqlite
from sqlalchemy.sql import func, select

from masakari import context
from masakari import db
from masakari.db.sqlalchemy import api as db_api
from masakari import test


class PurgeDeletedTest(test.TestCase):

    def setUp(self):
        super(PurgeDeletedTest, self).setUp()
        self.context = context.get_admin_context()
        self.engine = db_api.get_engine()
        self.notifications = sqlalchemyutils.get_table(
            self.engine, "notifications")
        self.failover_segments = sqlalchemyutils.get_table(
            self.engine, "failover_segments")
        # The hosts table has a FK of segment_id
        self.hosts = sqlalchemyutils.get_table(
            self.engine, "hosts")

        # Add 6 rows to table
        self.uuidstrs = []
        self.uuid_fs_segments = []
        self.uuid_hosts = []
        for record in range(6):
            notification_uuid = uuid.uuid4().hex
            fs_segment_uuid = uuid.uuid4().hex
            host_uuid = uuid.uuid4().hex
            ins_stmt = self.notifications.insert().values(
                notification_uuid=notification_uuid,
                generated_time=timeutils.utcnow(),
                source_host_uuid=host_uuid,
                type='demo',
                status='failed')
            self.uuidstrs.append(notification_uuid)
            with self.engine.connect() as conn, conn.begin():
                conn.execute(ins_stmt)

            ins_stmt = self.failover_segments.insert().values(
                uuid=fs_segment_uuid,
                name='test',
                service_type='demo',
                recovery_method='auto')
            self.uuid_fs_segments.append(fs_segment_uuid)
            with self.engine.connect() as conn, conn.begin():
                conn.execute(ins_stmt)

            ins_stmt = self.hosts.insert().values(
                uuid=host_uuid,
                failover_segment_id=fs_segment_uuid,
                name='host1',
                type='demo',
                control_attributes='test')
            self.uuid_hosts.append(host_uuid)
            with self.engine.connect() as conn, conn.begin():
                conn.execute(ins_stmt)

        # Set 4 of them deleted, 2 are 60 days ago, 2 are 20 days ago
        self.age_in_days_20 = timeutils.utcnow() - datetime.timedelta(days=20)
        self.age_in_days_60 = timeutils.utcnow() - datetime.timedelta(days=60)

        make_notifications_old = self.notifications.update().where(
            self.notifications.c.notification_uuid.in_(
                self.uuidstrs[1:3])).values(updated_at=self.age_in_days_20)
        make_notifications_older = self.notifications.update().where(
            self.notifications.c.notification_uuid.in_(
                self.uuidstrs[4:6])).values(updated_at=self.age_in_days_60)
        make_failover_segments_old = self.failover_segments.update().where(
            self.failover_segments.c.uuid.in_(
                self.uuid_fs_segments[1:3])).values(
            deleted_at=self.age_in_days_20)
        make_failover_segments_older = self.failover_segments.update().where(
            self.failover_segments.c.uuid.in_(
                self.uuid_fs_segments[4:6])).values(
            deleted_at=self.age_in_days_60)
        make_hosts_old = self.hosts.update().where(
            self.hosts.c.uuid.in_(self.uuid_hosts[1:3])).values(
            deleted_at=self.age_in_days_20)
        make_hosts_older = self.hosts.update().where(
            self.hosts.c.uuid.in_(self.uuid_hosts[4:6])).values(
            deleted_at=self.age_in_days_60)

        with self.engine.connect() as conn, conn.begin():
            conn.execute(make_notifications_old)
            conn.execute(make_notifications_older)
            conn.execute(make_failover_segments_old)
            conn.execute(make_failover_segments_older)
            conn.execute(make_hosts_old)
            conn.execute(make_hosts_older)

        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            with self.engine.connect() as conn, conn.begin():
                conn.exec_driver_sql("PRAGMA foreign_keys = ON")

    def _count(self, table):
        with self.engine.connect() as conn:
            return conn.execute(
                select(func.count()).select_from(table),
            ).scalar()

    def test_purge_deleted_rows_old(self):
        # Purge at 30 days old, should only delete 2 rows
        db.purge_deleted_rows(self.context, age_in_days=30, max_rows=10)

        notifications_rows = self._count(self.notifications)
        failover_segments_rows = self._count(self.failover_segments)
        hosts_rows = self._count(self.hosts)

        # Verify that we only deleted 2
        self.assertEqual(4, notifications_rows)
        self.assertEqual(4, failover_segments_rows)
        self.assertEqual(4, hosts_rows)

    def test_purge_all_deleted_rows(self):
        db.purge_deleted_rows(self.context, age_in_days=20, max_rows=-1)

        notifications_rows = self._count(self.notifications)
        failover_segments_rows = self._count(self.failover_segments)
        hosts_rows = self._count(self.hosts)

        # Verify that we have purged all deleted rows
        self.assertEqual(2, notifications_rows)
        self.assertEqual(2, failover_segments_rows)
        self.assertEqual(2, hosts_rows)

    def test_purge_maximum_rows_partial_deleted_records(self):
        notifications_rows = self._count(self.notifications)
        failover_segments_rows = self._count(self.failover_segments)
        hosts_rows = self._count(self.hosts)

        # Verify the initial number of rows
        self.assertEqual(6, notifications_rows)
        self.assertEqual(6, hosts_rows)
        self.assertEqual(6, failover_segments_rows)

        db.purge_deleted_rows(self.context, age_in_days=60, max_rows=3)

        notifications_rows = self._count(self.notifications)
        failover_segments_rows = self._count(self.failover_segments)
        hosts_rows = self._count(self.hosts)

        # Verify that we have deleted 3 rows only
        self.assertEqual(5, notifications_rows)
        self.assertEqual(4, hosts_rows)
        self.assertEqual(6, failover_segments_rows)



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
import re

import fixtures
from jsonschema import exceptions as jsonschema_exc

from masakari.api import api_version_request as api_version
from masakari.api import validation
from masakari.api.validation import parameter_types
from masakari.api.validation import validators
from masakari import exception
from masakari import test


class FakeRequest(object):
    def __init__(self, version=None):
        if version is None:
            version = '1.0'
        self.api_version_request = api_version.APIVersionRequest(version)


class ValidationRegex(test.NoDBTestCase):

    def test_build_regex_range(self):

        def _get_all_chars():
            for i in range(0x7F):
                yield chr(i)

        self.useFixture(fixtures.MonkeyPatch(
            'masakari.api.validation.parameter_types._get_all_chars',
            _get_all_chars))

        r = parameter_types._build_regex_range(ws=False)
        self.assertEqual(r, re.escape('!') + '-' + re.escape('~'))

        # if we allow whitespace the range starts earlier
        r = parameter_types._build_regex_range(ws=True)
        self.assertEqual(r, re.escape(' ') + '-' + re.escape('~'))

        # excluding a character will give us 2 ranges
        r = parameter_types._build_regex_range(ws=True, exclude=['A'])
        self.assertEqual(r,
                         re.escape(' ') + '-' + re.escape('@') +
                         'B' + '-' + re.escape('~'))

        # inverting which gives us all the initial unprintable characters.
        r = parameter_types._build_regex_range(ws=False, invert=True)
        self.assertEqual(r,
                         re.escape('\x00') + '-' + re.escape(' '))

        # excluding characters that create a singleton. Naively this would be:
        # ' -@B-BD-~' which seems to work, but ' -@BD-~' is more natural.
        r = parameter_types._build_regex_range(ws=True, exclude=['A', 'C'])
        self.assertEqual(r,
                         re.escape(' ') + '-' + re.escape('@') +
                         'B' + 'D' + '-' + re.escape('~'))

        # ws=True means the positive regex has printable whitespaces,
        # so the inverse will not. The inverse will include things we
        # exclude.
        r = parameter_types._build_regex_range(
            ws=True, exclude=['A', 'B', 'C', 'Z'], invert=True)
        self.assertEqual(r,
                         re.escape('\x00') + '-' + re.escape('\x1f') + 'A-CZ')


class APIValidationTestCase(test.NoDBTestCase):

    def setUp(self, schema=None):
        super(APIValidationTestCase, self).setUp()
        self.post = None

        if schema is not None:
            @validation.schema(request_body_schema=schema)
            def post(req, body):
                return 'Validation succeeded.'

            self.post = post

    def check_validation_error(self, method, body, expected_detail, req=None):
        if not req:
            req = FakeRequest()
        try:
            method(body=body, req=req,)
        except exception.ValidationError as ex:
            self.assertEqual(HTTPStatus.BAD_REQUEST, ex.kwargs['code'])
            if isinstance(expected_detail, list):
                self.assertIn(ex.kwargs['detail'], expected_detail,
                              'Exception details did not match expected')
            elif not re.match(expected_detail, ex.kwargs['detail']):
                self.assertEqual(expected_detail, ex.kwargs['detail'],
                                 'Exception details did not match expected')
        except Exception as ex:
            self.fail('An unexpected exception happens: %s' % ex)
        else:
            self.fail('Any exception does not happen.')


class FormatCheckerTestCase(test.NoDBTestCase):

    def test_format_checker_failed(self):
        format_checker = validators.FormatChecker()
        exc = self.assertRaises(jsonschema_exc.FormatError,
                                format_checker.check, "   ", "name")
        self.assertIsInstance(exc.cause, exception.InvalidName)
        self.assertEqual("An invalid 'name' value was provided. The name must "
                         "be: printable characters. "
                         "Can not start or end with whitespace.",
                         exc.cause.format_message())

    def test_format_checker_failed_with_non_string(self):
        checks = ["name"]
        format_checker = validators.FormatChecker()

        for check in checks:
            exc = self.assertRaises(jsonschema_exc.FormatError,
                                    format_checker.check, None, "name")
            self.assertIsInstance(exc.cause, exception.InvalidName)
            self.assertEqual("An invalid 'name' value was provided. The name "
                             "must be: printable characters. "
                             "Can not start or end with whitespace.",
                             exc.cause.format_message())


class RequiredDisableTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
        }
        super(RequiredDisableTestCase, self).setUp(schema=schema)

    def test_validate_required_disable(self):
        self.assertEqual(self.post(body={'foo': 1}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'abc': 1}, req=FakeRequest()),
                         'Validation succeeded.')


class RequiredEnableTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
            'required': ['foo']
        }
        super(RequiredEnableTestCase, self).setUp(schema=schema)

    def test_validate_required_enable(self):
        self.assertEqual(self.post(body={'foo': 1},
                                   req=FakeRequest()), 'Validation succeeded.')

    def test_validate_required_enable_fails(self):
        detail = "'foo' is a required property"
        self.check_validation_error(self.post, body={'abc': 1},
                                    expected_detail=detail)


class AdditionalPropertiesEnableTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
            'required': ['foo'],
        }
        super(AdditionalPropertiesEnableTestCase, self).setUp(schema=schema)

    def test_validate_additionalProperties_enable(self):
        self.assertEqual(self.post(body={'foo': 1}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': 1, 'ext': 1},
                                   req=FakeRequest()),
                         'Validation succeeded.')


class AdditionalPropertiesDisableTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'integer',
                },
            },
            'required': ['foo'],
            'additionalProperties': False,
        }
        super(AdditionalPropertiesDisableTestCase, self).setUp(schema=schema)

    def test_validate_additionalProperties_disable(self):
        self.assertEqual(self.post(body={'foo': 1}, req=FakeRequest()),
                         'Validation succeeded.')

    def test_validate_additionalProperties_disable_fails(self):
        detail = "Additional properties are not allowed ('ext' was unexpected)"
        self.check_validation_error(self.post, body={'foo': 1, 'ext': 1},
                                    expected_detail=detail)


class PatternPropertiesTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'patternProperties': {
                '^[a-zA-Z0-9]{1,10}$': {
                    'type': 'string'
                },
            },
            'additionalProperties': False,
        }
        super(PatternPropertiesTestCase, self).setUp(schema=schema)

    def test_validate_patternProperties(self):
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': 'bar'}, req=FakeRequest()))

    def test_validate_patternProperties_fails(self):
        details = [
            "Additional properties are not allowed ('__' was unexpected)",
            "'__' does not match any of the regexes: '^[a-zA-Z0-9]{1,10}$'"
        ]
        self.check_validation_error(self.post, body={'__': 'bar'},
                                    expected_detail=details)

        details = [
            "'' does not match any of the regexes: '^[a-zA-Z0-9]{1,10}$'",
            "Additional properties are not allowed ('' was unexpected)"
        ]
        self.check_validation_error(self.post, body={'': 'bar'},
                                    expected_detail=details)

        details = [
            ("'0123456789a' does not match any of the regexes: "
             "'^[a-zA-Z0-9]{1,10}$'"),
            ("Additional properties are not allowed ('0123456789a' was"
             " unexpected)")
        ]
        self.check_validation_error(self.post, body={'0123456789a': 'bar'},
                                    expected_detail=details)

        detail = "expected string or bytes-like object"
        self.check_validation_error(self.post, body={None: 'bar'},
                                    expected_detail=detail)


class StringTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'string',
                },
            },
        }
        super(StringTestCase, self).setUp(schema=schema)

    def test_validate_string(self):
        self.assertEqual(self.post(body={'foo': 'abc'}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': '0'}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': ''}, req=FakeRequest()),
                         'Validation succeeded.')

    def test_validate_string_fails(self):
        detail = ("Invalid input for field/attribute foo. Value: 1."
                  " 1 is not of type 'string'")
        self.check_validation_error(self.post, body={'foo': 1},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 1.5."
                  " 1.5 is not of type 'string'")
        self.check_validation_error(self.post, body={'foo': 1.5},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: True."
                  " True is not of type 'string'")
        self.check_validation_error(self.post, body={'foo': True},
                                    expected_detail=detail)


class StringLengthTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'string',
                    'minLength': 1,
                    'maxLength': 10,
                },
            },
        }
        super(StringLengthTestCase, self).setUp(schema=schema)

    def test_validate_string_length(self):
        self.assertEqual(self.post(body={'foo': '0'}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': '0123456789'},
                                   req=FakeRequest()),
                         'Validation succeeded.')

    def test_validate_string_length_fails(self):
        detail = [
            "Invalid input for field/attribute foo. Value: . '' "
            "is too short",  # jsonschema < 4.23.0
            "Invalid input for field/attribute foo. Value: . '' "
            "should be non-empty",  # jsonschema >= 4.23.0
        ]
        self.check_validation_error(self.post, body={'foo': ''},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 0123456789a."
                  " '0123456789a' is too long")
        self.check_validation_error(self.post, body={'foo': '0123456789a'},
                                    expected_detail=detail)


class IntegerTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': ['integer', 'string'],
                    'pattern': '^[0-9]+$',
                },
            },
        }
        super(IntegerTestCase, self).setUp(schema=schema)

    def test_validate_integer(self):
        self.assertEqual(self.post(body={'foo': 1}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': '1'}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': '0123456789'},
                                   req=FakeRequest()),
                         'Validation succeeded.')

    def test_validate_integer_fails(self):
        detail = ("Invalid input for field/attribute foo. Value: abc."
                  " 'abc' does not match '^[0-9]+$'")
        self.check_validation_error(self.post, body={'foo': 'abc'},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: True."
                  " True is not of type 'integer', 'string'")
        self.check_validation_error(self.post, body={'foo': True},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 0xffff."
                  " '0xffff' does not match '^[0-9]+$'")
        self.check_validation_error(self.post, body={'foo': '0xffff'},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 1.0."
                  " 1.0 is not of type 'integer', 'string'")
        self.check_validation_error(self.post, body={'foo': 1.0},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 1.0."
                  " '1.0' does not match '^[0-9]+$'")
        self.check_validation_error(self.post, body={'foo': '1.0'},
                                    expected_detail=detail)


class IntegerRangeTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': ['integer', 'string'],
                    'pattern': '^[0-9]+$',
                    'minimum': 1,
                    'maximum': 10,
                },
            },
        }
        super(IntegerRangeTestCase, self).setUp(schema=schema)

    def test_validate_integer_range(self):
        self.assertEqual(self.post(body={'foo': 1}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': 10}, req=FakeRequest()),
                         'Validation succeeded.')
        self.assertEqual(self.post(body={'foo': '1'}, req=FakeRequest()),
                         'Validation succeeded.')

    def test_validate_integer_range_fails(self):
        detail = ("Invalid input for field/attribute foo. Value: 0."
                  " 0(.0)? is less than the minimum of 1")
        self.check_validation_error(self.post, body={'foo': 0},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 11."
                  " 11(.0)? is greater than the maximum of 10")
        self.check_validation_error(self.post, body={'foo': 11},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 0."
                  " 0(.0)? is less than the minimum of 1")
        self.check_validation_error(self.post, body={'foo': '0'},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 11."
                  " 11(.0)? is greater than the maximum of 10")
        self.check_validation_error(self.post, body={'foo': '11'},
                                    expected_detail=detail)


class BooleanTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': parameter_types.boolean,
            },
        }
        super(BooleanTestCase, self).setUp(schema=schema)

    def test_validate_boolean(self):
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': True}, req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': False}, req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': 'True'}, req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': 'False'}, req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': '1'}, req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': '0'}, req=FakeRequest()))

    def test_validate_boolean_fails(self):
        enum_boolean = ("[True, 'True', 'TRUE', 'true', '1', 'ON', 'On',"
                        " 'on', 'YES', 'Yes', 'yes',"
                        " False, 'False', 'FALSE', 'false', '0', 'OFF', 'Off',"
                        " 'off', 'NO', 'No', 'no']")

        detail = ("Invalid input for field/attribute foo. Value: bar."
                  " 'bar' is not one of %s") % enum_boolean
        self.check_validation_error(self.post, body={'foo': 'bar'},
                                    expected_detail=detail)

        detail = ("Invalid input for field/attribute foo. Value: 2."
                  " '2' is not one of %s") % enum_boolean
        self.check_validation_error(self.post, body={'foo': '2'},
                                    expected_detail=detail)


class NameTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': parameter_types.name,
            },
        }
        super(NameTestCase, self).setUp(schema=schema)

    def test_validate_name(self):
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': 'm1.small'},
                                   req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': 'my server'},
                                   req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': 'a'}, req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': '\u0434'}, req=FakeRequest()))
        self.assertEqual('Validation succeeded.',
                         self.post(body={'foo': '\u0434\u2006\ufffd'},
                                   req=FakeRequest()))

    def test_validate_name_fails(self):
        error = ("An invalid 'name' value was provided. The name must be: "
                 "printable characters. "
                 "Can not start or end with whitespace.")

        should_fail = (' ',
                       ' segment',
                       'segment ',
                       'a\xa0',  # trailing unicode space
                       '\uffff',  # non-printable unicode
                       )

        for item in should_fail:
            self.check_validation_error(self.post, body={'foo': item},
                                    expected_detail=error)

        # four-byte unicode, if supported by this python build
        try:
            self.check_validation_error(self.post, body={'foo': '\U00010000'},
                                        expected_detail=error)
        except ValueError:
            pass


class DatetimeTestCase(APIValidationTestCase):

    def setUp(self):
        schema = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'string',
                    'format': 'date-time',
                },
            },
        }
        super(DatetimeTestCase, self).setUp(schema=schema)

    def test_validate_datetime(self):
        self.assertEqual('Validation succeeded.',
                         self.post(body={
                             'foo': '2016-01-14T01:00:00Z'}, req=FakeRequest()
                         ))


class VersionedApiValidationTestCase(APIValidationTestCase):

    def setUp(self):
        super(__class__, self).setUp()

        schema_pre13 = {
            'type': 'object',
            'properties': {
                'foo': {
                    'type': 'string',
                },
            },
            'additionalProperties': False,
        }

        schema_post13 = {
            'type': 'object',
            'properties': {
                'bar': {
                    'type': 'boolean',
                },
            },
            'additionalProperties': False,
        }

        @validation.schema(request_body_schema=schema_pre13,
                           min_version='1.1',
                           max_version='1.2')
        @validation.schema(request_body_schema=schema_post13,
                           min_version='1.3')
        def post(req, body):
            return 'Validation succeeded.'

        self.post = post

    def check_validation_error(self, body, req):
        try:
            self.post(body=body, req=req)
        except exception.ValidationError as ex:
            self.assertEqual(HTTPStatus.BAD_REQUEST, ex.kwargs['code'])
        except Exception as ex:
            self.fail('An unexpected exception happens: %s' % ex)
        else:
            self.fail('Any exception does not happen.')

    def test_validate_with_proper_microversions(self):
        self.assertEqual('Validation succeeded.',
                         self.post(body={
                             'foo': 'ahappystring'}, req=FakeRequest('1.1')
                         ))
        self.assertEqual('Validation succeeded.',
                         self.post(body={
                             'foo': 'ahappystring'}, req=FakeRequest('1.2')
                         ))
        self.assertEqual('Validation succeeded.',
                         self.post(body={
                             'bar': True}, req=FakeRequest('1.3')
                         ))
        self.assertEqual('Validation succeeded.',
                         self.post(body={
                             'bar': True}, req=FakeRequest('1.10')
                         ))
        self.assertEqual('Validation succeeded.',
                         self.post(body={
                             'whatever': None}, req=FakeRequest('1.0')
                         ))

    def test_validate_with_improper_microversions(self):
        self.check_validation_error(body={'bar': False},
                                    req=FakeRequest('1.1'))
        self.check_validation_error(body={'bar': False},
                                    req=FakeRequest('1.2'))
        self.check_validation_error(body={'foo': 'asadstring'},
                                    req=FakeRequest('1.3'))
        self.check_validation_error(body={'foo': 'asadstring'},
                                    req=FakeRequest('1.10'))
        self.check_validation_error(body={'foo': 'asadstring'},
                                    req=FakeRequest('2.0'))



# Copyright 2016 NTT DATA.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from masakari.api.validation import parameter_types
from masakari.objects import fields

create = {
    'type': 'object',
    'properties': {
        'notification': {
            'type': 'object',
            'properties': {
                'type': {
                    'type': 'string',
                    'enum': fields.NotificationType.ALL,
                },
                'hostname': parameter_types.hostname,
                'generated_time': {
                    'type': 'string',
                    'format': 'date-time',
                },
                'payload': parameter_types.payload,
            },
            'required': ['type', 'hostname', 'generated_time', 'payload'],
            'additionalProperties': False
        }
    },
    'required': ['notification'],
    'additionalProperties': False
}



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions that won't produce circular imports."""


def get_wrapped_function(function):
    """Get the method at the bottom of a stack of decorators."""
    if not hasattr(function, '__closure__') or not function.__closure__:
        return function

    def _get_wrapped_function(function):
        if not hasattr(function, '__closure__') or not function.__closure__:
            return None

        for closure in function.__closure__:
            func = closure.cell_contents

            deeper_func = _get_wrapped_function(func)
            if deeper_func:
                return deeper_func
            elif hasattr(closure.cell_contents, '__call__'):
                return closure.cell_contents

    return _get_wrapped_function(function)



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


auth_opts = [
    cfg.StrOpt("auth_strategy",
               default="keystone",
               choices=("keystone", "noauth2"),
               help="""
This determines the strategy to use for authentication: keystone or noauth2.
'noauth2' is designed for testing only, as it does no actual credential
checking. 'noauth2' provides administrative credentials only if 'admin' is
specified as the username.

* Possible values:

  Either 'keystone' (default) or 'noauth2'.

* Services that use this:

  ``masakari-api``

* Related options:

  None
"""),
    cfg.BoolOpt("use_forwarded_for",
                default=False,
                help="""
When True, the 'X-Forwarded-For' header is treated as the canonical remote
address. When False (the default), the 'remote_address' header is used.

You should only enable this if you have an HTML sanitizing proxy.

* Possible values:

  True, False (default)

* Services that use this:

  ``masakari-api``

* Related options:

  None
"""),
]


osapi_opts = [
    cfg.IntOpt("osapi_max_limit",
               default=1000,
               help="""
As a query can potentially return many thousands of items, you can limit the
maximum number of items in a single response by setting this option.

* Possible values:

  Any positive integer. Default is 1000.

* Services that use this:

  ``masakari-api``

* Related options:

  None
"""),
    cfg.StrOpt("osapi_masakari_link_prefix",
               help="""
This string is prepended to the normal URL that is returned in links to the
OpenStack Masakari API. If it is empty (the default), the URLs are returned
unchanged.

* Possible values:

  Any string, including an empty string (the default).

* Services that use this:

  ``masakari-api``

* Related options:

  None
"""),
]


ALL_OPTS = (auth_opts + osapi_opts)


def register_opts(conf):
    conf.register_opts(ALL_OPTS)


def list_opts():
    return {"DEFAULT": ALL_OPTS}



# Copyright 2016 NTT DATA
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
import inspect

from webob.util import status_reasons

from masakari import exception
from masakari import test


class MasakariExceptionTestCase(test.NoDBTestCase):
    def test_default_error_msg(self):
        class FakeMasakariException(exception.MasakariException):
            msg_fmt = "default message"

        exc = FakeMasakariException()
        self.assertEqual('default message', str(exc))

    def test_error_msg(self):
        self.assertEqual('test',
                         str(exception.MasakariException('test')))

    def test_default_error_msg_with_kwargs(self):
        class FakeMasakariException(exception.MasakariException):
            msg_fmt = "default message: %(code)s"

        exc = FakeMasakariException(code=int(HTTPStatus.INTERNAL_SERVER_ERROR))
        self.assertEqual('default message: 500', str(exc))
        self.assertEqual('default message: 500', exc.message)

    def test_error_msg_exception_with_kwargs(self):
        class FakeMasakariException(exception.MasakariException):
            msg_fmt = "default message: %(misspelled_code)s"

        exc = FakeMasakariException(code=int(HTTPStatus.INTERNAL_SERVER_ERROR),
                                    misspelled_code='blah')
        self.assertEqual('default message: blah', str(exc))
        self.assertEqual('default message: blah', exc.message)

    def test_default_error_code(self):
        class FakeMasakariException(exception.MasakariException):
            code = HTTPStatus.NOT_FOUND

        exc = FakeMasakariException()
        self.assertEqual(HTTPStatus.NOT_FOUND, exc.kwargs['code'])

    def test_error_code_from_kwarg(self):
        class FakeMasakariException(exception.MasakariException):
            code = HTTPStatus.INTERNAL_SERVER_ERROR

        exc = FakeMasakariException(code=HTTPStatus.NOT_FOUND)
        self.assertEqual(exc.kwargs['code'], HTTPStatus.NOT_FOUND)

    def test_format_message_local(self):
        class FakeMasakariException(exception.MasakariException):
            msg_fmt = "some message"

        exc = FakeMasakariException()
        self.assertEqual(str(exc), exc.format_message())

    def test_format_message_remote(self):
        class FakeMasakariException_Remote(exception.MasakariException):
            msg_fmt = "some message"

            def __str__(self):
                return "print the whole trace"

        exc = FakeMasakariException_Remote()
        self.assertEqual(u"print the whole trace", str(exc))
        self.assertEqual("some message", exc.format_message())

    def test_format_message_remote_error(self):
        class FakeMasakariException_Remote(exception.MasakariException):
            msg_fmt = "some message %(somearg)s"

            def __unicode__(self):
                return u"print the whole trace"

        self.flags(fatal_exception_format_errors=False)
        exc = FakeMasakariException_Remote(lame_arg='lame')
        self.assertEqual("some message %(somearg)s", exc.format_message())


class ConvertedExceptionTestCase(test.NoDBTestCase):
    def test_instantiate(self):
        exc = exception.ConvertedException(int(HTTPStatus.BAD_REQUEST),
                                           'Bad Request', 'reason')
        self.assertEqual(exc.code, HTTPStatus.BAD_REQUEST)
        self.assertEqual(exc.title, 'Bad Request')
        self.assertEqual(exc.explanation, 'reason')

    def test_instantiate_without_title_known_code(self):
        exc = exception.ConvertedException(
            int(HTTPStatus.INTERNAL_SERVER_ERROR))
        self.assertEqual(exc.title,
                         status_reasons[HTTPStatus.INTERNAL_SERVER_ERROR])

    def test_instantiate_without_title_unknown_code(self):
        exc = exception.ConvertedException(499)
        self.assertEqual(exc.title, 'Unknown Client Error')

    def test_instantiate_bad_code(self):
        self.assertRaises(KeyError, exception.ConvertedException, 10)


class ExceptionTestCase(test.NoDBTestCase):
    @staticmethod
    def _raise_exc(exc):
        raise exc(int(HTTPStatus.INTERNAL_SERVER_ERROR))

    def test_exceptions_raise(self):
        # NOTE(Dinesh_Bhor): disable format errors since we are not passing
        # kwargs
        self.flags(fatal_exception_format_errors=False)
        for name in dir(exception):
            exc = getattr(exception, name)
            # NOTE(yoctozepto): we skip HTTPStatus as it is not an exception
            # but a type also present in that module.
            if isinstance(exc, type) and name != 'HTTPStatus':
                self.assertRaises(exc, self._raise_exc, exc)


class ExceptionValidMessageTestCase(test.NoDBTestCase):

    def test_messages(self):
        failures = []

        for name, obj in inspect.getmembers(exception):
            if name in ['MasakariException', 'InstanceFaultRollback']:
                continue

            if not inspect.isclass(obj):
                continue

            if not issubclass(obj, exception.MasakariException):
                continue

            e = obj
            if e.msg_fmt == "An unknown exception occurred.":
                failures.append('%s needs a more specific msg_fmt' % name)

        if failures:
            self.fail('\n'.join(failures))



# Copyright (c) 2016 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


policy_data = """
{
    "context_is_admin": "role:admin or role:administrator",

    "os_masakari_api:extensions:index": "",
    "os_masakari_api:extensions:detail": "",
    "os_masakari_api:segments:index": "",
    "os_masakari_api:segments:detail": "",
    "os_masakari_api:segments:create": "",
    "os_masakari_api:segments:update": "",
    "os_masakari_api:segments:delete": "",
    "os_masakari_api:os-hosts:index": "",
    "os_masakari_api:os-hosts:detail": "",
    "os_masakari_api:os-hosts:create": "",
    "os_masakari_api:os-hosts:update": "",
    "os_masakari_api:os-hosts:delete": "",
    "os_masakari_api:notifications:index": "",
    "os_masakari_api:notifications:detail": "",
    "os_masakari_api:notifications:create": ""
}
"""



# Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log

from masakari.common import config
import masakari.conf
from masakari.db.sqlalchemy import api as sqlalchemy_api
from masakari import rpc
from masakari import version


CONF = masakari.conf.CONF


def parse_args(argv, default_config_files=None, configure_db=True,
               init_rpc=True):
    log.register_options(CONF)
    # We use the oslo.log default log levels which includes suds=INFO
    # and add only the extra levels that Masakari needs
    log.set_defaults(default_log_levels=log.get_default_log_levels())
    rpc.set_defaults(control_exchange='masakari')
    config.set_middleware_defaults()

    CONF(argv[1:],
         project='masakari',
         version=version.version_string(),
         default_config_files=default_config_files)

    if init_rpc:
        rpc.init(CONF)

    if configure_db:
        sqlalchemy_api.configure(CONF)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstack import exceptions

from masakari.objects import fields
from masakari.tests.functional import base


class TestSegments(base.BaseFunctionalTest):
    def test_create_get_delete(self):
        # This test will create, get and delete a segment
        segment_data = {
            'name': self.getUniqueString(),
            'recovery_method': fields.FailoverSegmentRecoveryMethod.AUTO,
            'service_type': 'COMPUTE',
        }
        segment = self.admin_conn.ha.create_segment(**segment_data)

        self.assertEqual(dict(segment), {**segment_data, **segment})

        result = self.admin_conn.ha.get_segment(segment.uuid)

        self.assertEqual(segment.name, result.name)
        self.assertEqual(segment.recovery_method, result.recovery_method)
        self.assertEqual(segment.service_type, result.service_type)

        self.admin_conn.ha.delete_segment(segment.uuid)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.admin_conn.ha.get_segment, segment.uuid)

    def test_create_delete_with_host(self):
        # This test is for deleting a segment with hosts
        if not self.hypervisors:
            self.skipTest("Skipped as there are no hypervisors "
                          "configured in nova")

        segment = self.admin_conn.ha.create_segment(
            name=self.getUniqueString(),
            recovery_method=fields.FailoverSegmentRecoveryMethod.AUTO,
            service_type='COMPUTE')

        # Create valid host
        host_name = self.hypervisors[0]['name']

        host = self.admin_conn.ha.create_host(segment_id=segment.uuid,
                                        name=host_name,
                                        type='COMPUTE',
                                        control_attributes='SSH')

        result = self.admin_conn.ha.get_segment(segment.uuid)
        self.assertEqual(segment.name, result.name)

        # Delete segment, which should delete hosts as well
        self.admin_conn.ha.delete_segment(segment['uuid'])
        self.assertRaises(exceptions.ResourceNotFound,
                          self.admin_conn.ha.get_segment, segment.uuid)
        self.assertRaises(exceptions.ResourceNotFound,
                          self.admin_conn.ha.get_host, host.uuid, segment.uuid)

    def test_list(self):
        # This test is for listing segments using filters
        segment_data_1 = {'name': self.getUniqueString(),
                  'recovery_method': fields.FailoverSegmentRecoveryMethod.AUTO,
                  'service_type': 'COMPUTE'}

        segment_data_2 = {'name': self.getUniqueString(),
                  'recovery_method':
                      fields.FailoverSegmentRecoveryMethod.RESERVED_HOST,
                  'service_type': 'COMPUTE'}

        # Create segments
        segment_1 = self.admin_conn.ha.create_segment(**segment_data_1)
        segment_2 = self.admin_conn.ha.create_segment(**segment_data_2)

        # Delete segments
        self.addCleanup(self.admin_conn.ha.delete_segment, segment_1.uuid)
        self.addCleanup(self.admin_conn.ha.delete_segment, segment_2.uuid)

        segments = self.admin_conn.ha.segments()
        self.assertCountEqual([segment_1, segment_2], segments)

    def test_list_with_filter(self):
        # This test is for listing segments using filters
        segment_data_1 = {'name': self.getUniqueString(),
                  'recovery_method': fields.FailoverSegmentRecoveryMethod.AUTO,
                  'service_type': 'COMPUTE'}

        segment_data_2 = {'name': self.getUniqueString(),
                  'recovery_method':
                      fields.FailoverSegmentRecoveryMethod.RESERVED_HOST,
                  'service_type': 'COMPUTE'}

        # Create segments
        segment_1 = self.admin_conn.ha.create_segment(**segment_data_1)
        segment_2 = self.admin_conn.ha.create_segment(**segment_data_2)

        # Delete segments
        self.addCleanup(self.admin_conn.ha.delete_segment, segment_1.uuid)
        self.addCleanup(self.admin_conn.ha.delete_segment, segment_2.uuid)

        for seg_object in self.admin_conn.ha.segments(
                recovery_method=fields.FailoverSegmentRecoveryMethod.AUTO):

            self.assertEqual(
                dict(seg_object), {**segment_data_1, **seg_object})

        for seg_object in self.admin_conn.ha.segments(
                recovery_method=fields.FailoverSegmentRecoveryMethod.
                RESERVED_HOST):

            self.assertEqual(
                dict(seg_object), {**segment_data_2, **seg_object})

    def test_update_with_host(self):
        # This test is for updating segment with host
        if not self.hypervisors:
            self.skipTest("Skipped as there are no hypervisors "
                          "configured in nova")

        segment = self.admin_conn.ha.create_segment(
            name=self.getUniqueString(),
            recovery_method=fields.FailoverSegmentRecoveryMethod.AUTO,
            service_type='COMPUTE')

        # Delete segment
        self.addCleanup(self.admin_conn.ha.delete_segment, segment.uuid)

        # Create valid host
        host_name = self.hypervisors[0]['name']

        self.admin_conn.ha.create_host(segment_id=segment.uuid, name=host_name,
                                type='COMPUTE', control_attributes='SSH')

        # Update segment
        segment_1 = self.admin_conn.ha.update_segment(segment.uuid,
            name=self.getUniqueString(),
            recovery_method=fields.FailoverSegmentRecoveryMethod.RESERVED_HOST,
            service_type='CONTROLLER')

        result = self.admin_conn.ha.get_segment(segment.uuid)
        self.assertEqual(segment_1.name, result.name)
        self.assertEqual(segment_1.recovery_method, result.recovery_method)
        self.assertEqual(segment_1.service_type, result.service_type)



# Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from masakari import exception
from masakari.i18n import _

# Define the minimum and maximum version of the API across all of the
# REST API. The format of the version is:
# X.Y where:
#
# - X will only be changed if a significant backwards incompatible API
# change is made which affects the API as whole. That is, something
# that is only very very rarely incremented.
#
# - Y when you make any change to the API. Note that this includes
# semantic changes which may not affect the input or output formats or
# even originate in the API code layer. We are not distinguishing
# between backwards compatible and backwards incompatible changes in
# the versioning system. It must be made clear in the documentation as
# to what is a backwards compatible change and what is a backwards
# incompatible one.

#
# You must update the API version history string below with a one or
# two line description as well as update rest_api_version_history.rst
REST_API_VERSION_HISTORY = """REST API Version History:

    * 1.0 - Initial version.
    * 1.1 - Add support for getting notification progress details.
    * 1.2 - Add enabled option to segment.
    * 1.3 - Add masakari vmoves.
"""

# The minimum and maximum versions of the API supported
# The default api version request is defined to be the
# the minimum version of the API supported.
# Note: This only applies for the v1 API once microversions
# support is fully merged.
_MIN_API_VERSION = "1.0"
_MAX_API_VERSION = "1.3"
# The default api version request if none is requested in the headers
DEFAULT_API_VERSION = _MIN_API_VERSION


# NOTE: min and max versions declared as functions so we can
# mock them for unittests. Do not use the constants directly anywhere
# else.
def min_api_version():
    return APIVersionRequest(_MIN_API_VERSION)


def max_api_version():
    return APIVersionRequest(_MAX_API_VERSION)


def is_supported(req, min_version=_MIN_API_VERSION,
                 max_version=_MAX_API_VERSION):
    """Check if API request version satisfies version restrictions.

    :param req: request object
    :param min_version: minimal version of API needed for correct
           request processing
    :param max_version: maximum version of API needed for correct
           request processing

    :returns True if request satisfies minimal and maximum API version
             requirements. False in other case.
    """

    return (APIVersionRequest(max_version) >= req.api_version_request >=
            APIVersionRequest(min_version))


class APIVersionRequest(object):
    """This class represents an API Version Request with convenience
    methods for manipulation and comparison of version
    numbers that we need to do to implement microversions.
    """

    def __init__(self, version_string=None):
        """Create an API version request object.

        :param version_string: String representation of APIVersionRequest.
            Correct format is 'X.Y', where 'X' and 'Y' are int values.
            None value should be used to create Null APIVersionRequest,
            which is equal to 0.0
        """
        self.ver_major = 0
        self.ver_minor = 0

        if version_string is not None:
            match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0)$",
                             version_string)
            if match:
                self.ver_major = int(match.group(1))
                self.ver_minor = int(match.group(2))
            else:
                raise exception.InvalidAPIVersionString(version=version_string)

    def __str__(self):
        """Debug/Logging representation of object."""
        return ("API Version Request Major: %s, Minor: %s"
                % (self.ver_major, self.ver_minor))

    def is_null(self):
        return self.ver_major == 0 and self.ver_minor == 0

    def _format_type_error(self, other):
        return TypeError(_("'%(other)s' should be an instance of '%(cls)s'") %
                         {"other": other, "cls": self.__class__})

    def __lt__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) <
                (other.ver_major, other.ver_minor))

    def __eq__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) ==
                (other.ver_major, other.ver_minor))

    def __gt__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) >
                (other.ver_major, other.ver_minor))

    def __le__(self, other):
        return self < other or self == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return self > other or self == other

    def matches(self, min_version, max_version):
        """Returns whether the version object represents a version
        greater than or equal to the minimum version and less than
        or equal to the maximum version.

        @param min_version: Minimum acceptable version.
        @param max_version: Maximum acceptable version.
        @returns: boolean

        If min_version is null then there is no minimum limit.
        If max_version is null then there is no maximum limit.
        If self is null then raise ValueError
        """

        if self.is_null():
            raise ValueError
        if max_version.is_null() and min_version.is_null():
            return True
        elif max_version.is_null():
            return min_version <= self
        elif min_version.is_null():
            return self <= max_version
        else:
            return min_version <= self <= max_version

    def get_string(self):
        """Converts object to string representation which if used to create
        an APIVersionRequest object results in the same version request.
        """
        if self.is_null():
            raise ValueError
        return "%s.%s" % (self.ver_major, self.ver_minor)



# Copyright (c) 2018 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from masakari.notifications.objects import base
from masakari.notifications.objects import exception
from masakari.notifications.objects import notification as event_notification
from masakari.objects import host as host_obj
from masakari.objects import notification as notification_obj
from masakari.objects import segment as segment_obj


segment = segment_obj.FailoverSegment()
host = host_obj.Host()
notification = notification_obj.Notification()
fault = None


init_args = {
    event_notification.SegmentApiPayloadBase: [segment],
    event_notification.SegmentApiPayload: [segment, fault],
    event_notification.HostApiPayloadBase: [host],
    event_notification.HostApiPayload: [host, fault],
    event_notification.NotificationApiPayloadBase: [notification],
    event_notification.NotificationApiPayload: [notification, fault],
    event_notification.SegmentApiNotification: [],
    event_notification.HostApiNotification: [],
    event_notification.NotificationApiNotification: [],
    exception.ExceptionPayload: [],
    exception.ExceptionNotification: [],
    base.EventType: [],
    base.NotificationPublisher: [],
    segment_obj.FailoverSegment: [],
    segment_obj.FailoverSegmentList: [],
    host_obj.Host: [],
    host_obj.HostList: [],
    notification_obj.Notification: [],
    notification_obj.NotificationList: [],
}


init_kwargs = {
    event_notification.SegmentApiPayloadBase: {},
    event_notification.SegmentApiPayload: {},
    event_notification.HostApiPayloadBase: {},
    event_notification.HostApiPayload: {},
    event_notification.NotificationApiPayloadBase: {},
    event_notification.NotificationApiPayload: {},
    event_notification.SegmentApiNotification: {},
    event_notification.HostApiNotification: {},
    event_notification.NotificationApiNotification: {},
    exception.ExceptionPayload: {},
    exception.ExceptionNotification: {},
    base.EventType: {},
    base.NotificationPublisher: {},
    segment_obj.FailoverSegment: {},
    segment_obj.FailoverSegmentList: {},
    host_obj.Host: {},
    host_obj.HostList: {},
    notification_obj.Notification: {},
    notification_obj.NotificationList: {},
}



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils
from oslo_utils import versionutils

from masakari.api import utils as api_utils
from masakari import db
from masakari import exception
from masakari import objects
from masakari.objects import base
from masakari.objects import fields

LOG = logging.getLogger(__name__)


@base.MasakariObjectRegistry.register
class FailoverSegment(base.MasakariPersistentObject, base.MasakariObject,
                      base.MasakariObjectDictCompat):
    # 1.0, init
    # 1.1, add enabled field
    VERSION = '1.1'

    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.UUIDField(),
        'name': fields.StringField(),
        'service_type': fields.StringField(),
        'enabled': fields.BooleanField(default=True),
        'description': fields.StringField(nullable=True),
        'recovery_method': fields.FailoverSegmentRecoveryMethodField(),
        }

    def obj_make_compatible(self, primitive, target_version):
        super(FailoverSegment, self).obj_make_compatible(primitive,
                                                         target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1) and 'enabled' in primitive:
            del primitive['enabled']

    @staticmethod
    def _from_db_object(context, segment, db_segment):
        for key in segment.fields:
            setattr(segment, key, db_segment[key])
        segment._context = context
        segment.obj_reset_changes()
        return segment

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        db_inst = db.failover_segment_get_by_id(context, id)
        return cls._from_db_object(context, cls(), db_inst)

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid):
        db_inst = db.failover_segment_get_by_uuid(context, uuid)
        return cls._from_db_object(context, cls(), db_inst)

    @base.remotable_classmethod
    def get_by_name(cls, context, name):
        db_inst = db.failover_segment_get_by_name(context, name)
        return cls._from_db_object(context, cls(), db_inst)

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason='already created')
        updates = self.masakari_obj_get_changes()

        if 'uuid' not in updates:
            updates['uuid'] = uuidutils.generate_uuid()
            LOG.debug('Generated uuid %(uuid)s for failover segment',
                      dict(uuid=updates['uuid']))

        api_utils.notify_about_segment_api(self._context, self,
            action=fields.EventNotificationAction.SEGMENT_CREATE,
            phase=fields.EventNotificationPhase.START)

        db_segment = db.failover_segment_create(self._context, updates)

        api_utils.notify_about_segment_api(self._context, self,
            action=fields.EventNotificationAction.SEGMENT_CREATE,
            phase=fields.EventNotificationPhase.END)

        self._from_db_object(self._context, self, db_segment)

    @base.remotable
    def save(self):
        updates = self.masakari_obj_get_changes()
        updates.pop('id', None)

        api_utils.notify_about_segment_api(self._context, self,
            action=fields.EventNotificationAction.SEGMENT_UPDATE,
            phase=fields.EventNotificationPhase.START)

        db_segment = db.failover_segment_update(self._context,
                                                self.uuid, updates)

        api_utils.notify_about_segment_api(self._context, self,
            action=fields.EventNotificationAction.SEGMENT_UPDATE,
            phase=fields.EventNotificationPhase.END)

        self._from_db_object(self._context, self, db_segment)

    @base.remotable
    def destroy(self):
        if not self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='destroy',
                                              reason='already destroyed')
        if not self.obj_attr_is_set('uuid'):
            raise exception.ObjectActionError(action='destroy',
                                              reason='no uuid')

        api_utils.notify_about_segment_api(self._context, self,
            action=fields.EventNotificationAction.SEGMENT_DELETE,
            phase=fields.EventNotificationPhase.START)

        db.failover_segment_delete(self._context, self.uuid)

        api_utils.notify_about_segment_api(self._context, self,
            action=fields.EventNotificationAction.SEGMENT_DELETE,
            phase=fields.EventNotificationPhase.END)

        delattr(self, base.get_attrname('id'))

    def is_under_recovery(self, filters=None):
        return db.is_failover_segment_under_recovery(self._context, self.uuid,
                                                     filters=filters)


@base.MasakariObjectRegistry.register
class FailoverSegmentList(base.ObjectListBase, base.MasakariObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('FailoverSegment'),
        }

    @base.remotable_classmethod
    def get_all(cls, ctxt, filters=None, sort_keys=None,
                sort_dirs=None, limit=None, marker=None):

        groups = db.failover_segment_get_all_by_filters(ctxt, filters=filters,
                                                        sort_keys=sort_keys,
                                                        sort_dirs=sort_dirs,
                                                        limit=limit,
                                                        marker=marker)

        return base.obj_make_list(ctxt, cls(ctxt), objects.FailoverSegment,
                                  groups)



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import models
from oslo_utils import timeutils
from sqlalchemy import (Column, DateTime, Index, Integer, Enum, String,
                        schema)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import orm
from sqlalchemy import ForeignKey, Boolean, Text


BASE = declarative_base()


class MasakariTimestampMixin(object):
    # Note(tpatil): timeutils.utcnow() method return microseconds part but db
    # doesn't store it because of which subsequent calls to get resources
    # from the same db session object instance doesn't return microsecond for
    # datetime fields. To avoid this discrepancy, removed microseconds from
    # datetime fields so that there is no need to remove it for create/update
    # cases in the respective versioned objects.
    created_at = Column(DateTime, default=lambda: timeutils.utcnow().replace(
                        microsecond=0))
    updated_at = Column(DateTime, onupdate=lambda: timeutils.utcnow().replace(
                        microsecond=0))


class MasakariAPIBase(MasakariTimestampMixin, models.ModelBase):
    """Base class for MasakariAPIBase Models."""

    metadata = None

    def __copy__(self):
        """Implement a safe copy.copy().

        SQLAlchemy-mapped objects travel with an object
        called an InstanceState, which is pegged to that object
        specifically and tracks everything about that object.  It's
        critical within all attribute operations, including gets
        and deferred loading.   This object definitely cannot be
        shared among two instances, and must be handled.

        The copy routine here makes use of session.merge() which
        already essentially implements a "copy" style of operation,
        which produces a new instance with a new InstanceState and copies
        all the data along mapped attributes without using any SQL.

        The mode we are using here has the caveat that the given object
        must be "clean", e.g. that it has no database-loaded state
        that has been updated and not flushed.   This is a good thing,
        as creating a copy of an object including non-flushed, pending
        database state is probably not a good idea; neither represents
        what the actual row looks like, and only one should be flushed.

        """
        session = orm.Session()

        copy = session.merge(self, load=False)
        session.expunge(copy)
        return copy


class FailoverSegment(BASE, MasakariAPIBase, models.SoftDeleteMixin):
    """Represents a failover segment."""
    __tablename__ = 'failover_segments'
    __table_args__ = (
        schema.UniqueConstraint("name", "deleted",
                                name="uniq_segment0name0deleted"),
        schema.UniqueConstraint('uuid', name='uniq_segments0uuid'),
        Index('segments_service_type_idx', 'service_type'),
    )

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(String(36), nullable=False)
    name = Column(String(255), nullable=False)
    service_type = Column(String(255), nullable=False)
    enabled = Column(Boolean, default=True)
    description = Column(Text)
    recovery_method = Column(Enum('auto', 'reserved_host', 'auto_priority',
                                  'rh_priority',
                                  name='recovery_methods'), nullable=False)


class Host(BASE, MasakariAPIBase, models.SoftDeleteMixin):
    """Represents a host."""
    __tablename__ = 'hosts'
    __table_args__ = (
        schema.UniqueConstraint("name", "deleted",
                                name="uniq_host0name0deleted"),
        schema.UniqueConstraint('uuid', name='uniq_host0uuid'),
        Index('hosts_type_idx', 'type'),
    )

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(String(36), nullable=False)
    name = Column(String(255), nullable=False)
    reserved = Column(Boolean, default=False)
    type = Column(String(255), nullable=False)
    control_attributes = Column(Text, nullable=False)
    on_maintenance = Column(Boolean, default=False)
    failover_segment_id = Column(String(36),
                                 ForeignKey('failover_segments.uuid'),
                                 nullable=False)

    failover_segment = orm.relationship(FailoverSegment,
                                        backref=orm.backref('hosts'),
                                        foreign_keys=failover_segment_id,
                                        primaryjoin='and_(Host.'
                                                    'failover_segment_id=='
                                                    'FailoverSegment.uuid,'
                                                    'Host.deleted==0)')


class Notification(BASE, MasakariAPIBase, models.SoftDeleteMixin):
    """Represents a notification."""
    __tablename__ = 'notifications'
    __table_args__ = (
        schema.UniqueConstraint('notification_uuid',
                                name='uniq_notification0uuid'),
    )

    id = Column(Integer, primary_key=True, autoincrement=True)
    notification_uuid = Column(String(36), nullable=False)
    generated_time = Column(DateTime, nullable=False)
    type = Column(String(36), nullable=False)
    payload = Column(Text)
    status = Column(Enum('new', 'running', 'error', 'failed',
                         'ignored', 'finished', name='notification_status'),
                    nullable=False)
    source_host_uuid = Column(String(36), nullable=False)


class VMove(BASE, MasakariAPIBase, models.SoftDeleteMixin):
    """Represents one vm move."""
    __tablename__ = 'vmoves'
    __table_args__ = (
        schema.UniqueConstraint('uuid',
                                name='uniq_vmove0uuid'),
    )

    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(String(36), nullable=False)
    notification_uuid = Column(String(36), nullable=False)
    instance_uuid = Column(String(36), nullable=False)
    instance_name = Column(String(255), nullable=False)
    source_host = Column(String(255), nullable=True)
    dest_host = Column(String(255), nullable=True)
    start_time = Column(DateTime, nullable=True)
    end_time = Column(DateTime, nullable=True)
    type = Column(String(36), nullable=True)
    status = Column(String(255), nullable=True)
    message = Column(Text)



#    Copyright (c) 2016 NTT DATA
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus

from oslo_middleware import request_id
from oslo_serialization import jsonutils
import webob
import webob.exc

import masakari.api.auth
import masakari.conf
from masakari import test

CONF = masakari.conf.CONF


class TestMasakariKeystoneContextMiddleware(test.NoDBTestCase):

    def setUp(self):
        super(TestMasakariKeystoneContextMiddleware, self).setUp()

        @webob.dec.wsgify()
        def fake_app(req):
            self.context = req.environ['masakari.context']
            return webob.Response()

        self.context = None
        self.middleware = masakari.api.auth.MasakariKeystoneContext(fake_app)
        self.request = webob.Request.blank('/')
        self.request.headers['X_TENANT_ID'] = 'testtenantid'
        self.request.headers['X_AUTH_TOKEN'] = 'testauthtoken'
        self.request.headers['X_SERVICE_CATALOG'] = jsonutils.dumps({})

    def test_no_user_or_user_id(self):
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.UNAUTHORIZED)

    def test_user_id_only(self):
        self.request.headers['X_USER_ID'] = 'testuserid'
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.OK)
        self.assertEqual(self.context.user_id, 'testuserid')

    def test_user_only(self):
        self.request.headers['X_USER'] = 'testuser'
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.OK)
        self.assertEqual(self.context.user_id, 'testuser')

    def test_user_id_trumps_user(self):
        self.request.headers['X_USER_ID'] = 'testuserid'
        self.request.headers['X_USER'] = 'testuser'
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.OK)
        self.assertEqual(self.context.user_id, 'testuserid')

    def test_invalid_service_catalog(self):
        self.request.headers['X_USER'] = 'testuser'
        self.request.headers['X_SERVICE_CATALOG'] = "bad json"
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.INTERNAL_SERVER_ERROR)

    def test_request_id_extracted_from_env(self):
        req_id = 'dummy-request-id'
        self.request.headers['X_PROJECT_ID'] = 'testtenantid'
        self.request.headers['X_USER_ID'] = 'testuserid'
        self.request.environ[request_id.ENV_REQUEST_ID] = req_id
        self.request.get_response(self.middleware)
        self.assertEqual(req_id, self.context.request_id)


class TestNoAuthMiddleware(test.NoDBTestCase):

    def setUp(self):
        super(TestNoAuthMiddleware, self).setUp()

        @webob.dec.wsgify()
        def fake_app(req):
            self.context = req.environ['masakari.context']
            return webob.Response()

        self.context = None
        self.middleware = masakari.api.auth.NoAuthMiddleware(fake_app)
        self.request = webob.Request.blank('/')
        self.request.headers['X_SERVICE_CATALOG'] = jsonutils.dumps({})

    def test_no_user_or_user_id(self):
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.OK)

    def test_user_id_only(self):
        self.request.headers['X_USER_ID'] = 'testuserid'
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.OK)
        self.assertEqual(self.context.user_id, 'testuserid')

    def test_user_only(self):
        self.request.headers['X_USER'] = 'testuser'
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.OK)
        self.assertEqual(self.context.user_id, 'testuser')

    def test_user_id_trumps_user(self):
        self.request.headers['X_USER_ID'] = 'testuserid'
        self.request.headers['X_USER'] = 'testuser'
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.OK)
        self.assertEqual(self.context.user_id, 'testuserid')

    def test_invalid_service_catalog(self):
        self.request.headers['X_USER'] = 'testuser'
        self.request.headers['X_SERVICE_CATALOG'] = "bad json"
        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status_int, HTTPStatus.INTERNAL_SERVER_ERROR)

    def test_request_id_extracted_from_env(self):
        req_id = 'dummy-request-id'
        self.request.headers['X_PROJECT_ID'] = 'testtenantid'
        self.request.headers['X_USER_ID'] = 'testuserid'
        self.request.environ[request_id.ENV_REQUEST_ID] = req_id
        self.request.get_response(self.middleware)
        self.assertEqual(req_id, self.context.request_id)


class TestKeystoneMiddlewareRoles(test.NoDBTestCase):

    def setUp(self):
        super(TestKeystoneMiddlewareRoles, self).setUp()

        @webob.dec.wsgify()
        def role_check_app(req):
            context = req.environ['masakari.context']

            if "knight" in context.roles and "bad" not in context.roles:
                return webob.Response(status="200 Role Match")
            elif not context.roles:
                return webob.Response(status="200 No Roles")
            else:
                raise webob.exc.HTTPBadRequest("unexpected role header")

        self.middleware = (
            masakari.api.auth.MasakariKeystoneContext(role_check_app))
        self.request = webob.Request.blank('/')
        self.request.headers['X_USER'] = 'testuser'
        self.request.headers['X_TENANT_ID'] = 'testtenantid'
        self.request.headers['X_AUTH_TOKEN'] = 'testauthtoken'
        self.request.headers['X_SERVICE_CATALOG'] = jsonutils.dumps({})

        self.roles = "pawn, knight, rook"

    def test_roles(self):
        self.request.headers['X_ROLES'] = 'pawn,knight,rook'

        response = self.request.get_response(self.middleware)
        self.assertEqual(response.status, '200 Role Match')


class TestPipeLineFactory(test.NoDBTestCase):

    class FakeFilter(object):
        def __init__(self, name):
            self.name = name
            self.obj = None

        def __call__(self, obj):
            self.obj = obj
            return self

    class FakeApp(object):
        def __init__(self, name):
            self.name = name

    class FakeLoader(object):
        def get_filter(self, name):
            return TestPipeLineFactory.FakeFilter(name)

        def get_app(self, name):
            return TestPipeLineFactory.FakeApp(name)

    def _test_pipeline(self, pipeline, app):
        for p in pipeline.split()[:-1]:
            self.assertEqual(app.name, p)
            self.assertIsInstance(app, TestPipeLineFactory.FakeFilter)
            app = app.obj
        self.assertEqual(app.name, pipeline.split()[-1])
        self.assertIsInstance(app, TestPipeLineFactory.FakeApp)

    def test_pipeline_factory_v1(self):
        fake_pipeline = 'test1 test2 test3'
        CONF.set_override('auth_strategy', 'noauth2')
        app = masakari.api.auth.pipeline_factory_v1(
            TestPipeLineFactory.FakeLoader(), None, noauth2=fake_pipeline)
        self._test_pipeline(fake_pipeline, app)



# Copyright 2016 NTT DATA
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from masakari.api.openstack.ha import extension_info
from masakari import policy
from masakari import test
from masakari.tests.unit.api.openstack import fakes


class fake_extension(object):
    def __init__(self, name, alias, description, version):
        self.name = name
        self.alias = alias
        self.__doc__ = description
        self.version = version


fake_extensions = {
    'ext1-alias': fake_extension('ext1', 'ext1-alias', 'ext1 description', 1),
    'ext2-alias': fake_extension('ext2', 'ext2-alias', 'ext2 description', 2),
    'ext3-alias': fake_extension('ext3', 'ext3-alias', 'ext3 description', 1)
}

simulated_extension_list = {
    'segments': fake_extension('Segments', 'segments', 'Segments.', 1),
    'hosts': fake_extension('Hosts', 'hosts', 'Hosts.', 2),
    'os-fake': fake_extension('Cells', 'os-fake', 'Cells description', 1)
}


def fake_policy_authorize_selective(context, action, target):
    return action != 'os_masakari_api:ext1-alias:discoverable'


class ExtensionInfoTest(test.NoDBTestCase):

    def setUp(self):
        super(ExtensionInfoTest, self).setUp()
        ext_info = extension_info.LoadedExtensionInfo()
        ext_info.extensions = fake_extensions
        self.controller = extension_info.ExtensionInfoController(ext_info)

    def _filter_extensions(self, res_dict):

        for e in [x for x in res_dict['extensions'] if '-alias' in x['alias']]:
            self.assertIn(e['alias'], fake_extensions)
            self.assertEqual(e['name'], fake_extensions[e['alias']].name)
            self.assertEqual(e['alias'], fake_extensions[e['alias']].alias)
            self.assertEqual(e['description'],
                             fake_extensions[e['alias']].__doc__)
            self.assertEqual(e['updated'], "")
            self.assertEqual(e['links'], [])
            self.assertEqual(6, len(e))

    @mock.patch.object(policy, 'authorize', mock.Mock(return_value=True))
    def test_extension_info_list(self):
        req = fakes.HTTPRequest.blank('/extensions')
        res_dict = self.controller.index(req)
        self.assertGreaterEqual(len(res_dict['extensions']), 3)
        self._filter_extensions(res_dict)

    @mock.patch.object(policy, 'authorize', mock.Mock(return_value=True))
    def test_extension_info_show(self):
        req = fakes.HTTPRequest.blank('/extensions/ext1-alias')
        res_dict = self.controller.show(req, 'ext1-alias')
        self.assertEqual(1, len(res_dict))
        self.assertEqual(res_dict['extension']['name'],
                         fake_extensions['ext1-alias'].name)
        self.assertEqual(res_dict['extension']['alias'],
                         fake_extensions['ext1-alias'].alias)
        self.assertEqual(res_dict['extension']['description'],
                         fake_extensions['ext1-alias'].__doc__)
        self.assertEqual(res_dict['extension']['updated'], "")
        self.assertEqual(res_dict['extension']['links'], [])
        self.assertEqual(6, len(res_dict['extension']))

    @mock.patch.object(policy, 'authorize')
    def test_extension_info_list_not_all_discoverable(self, mock_authorize):
        mock_authorize.side_effect = fake_policy_authorize_selective
        req = fakes.HTTPRequest.blank('/extensions')
        res_dict = self.controller.index(req)
        self.assertGreaterEqual(len(res_dict['extensions']), 2)
        self._filter_extensions(res_dict)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


coordination_opts = [
    cfg.StrOpt('backend_url',
               default=None,
               help="The backend URL to use for distributed coordination."
                    "By default it's None which means that coordination is "
                    "disabled. The coordination is implemented for "
                    "distributed lock management and was tested with etcd."
                    "Coordination doesn't work for file driver because lock "
                    "files aren't removed after lock releasing."),
]


def register_opts(conf):
    """Registers coordination configuration options
    :param conf: configuration
    """
    conf.register_opts(coordination_opts, group="coordination")


def list_opts():
    """Lists coordination configuration options
    :return: coordination configuration options
    """
    return {"coordination": coordination_opts}



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Driver TaskFlowDriver:

    Execute notification workflows using taskflow.
"""
from collections import OrderedDict
import contextlib

from oslo_log import log as logging
from oslo_utils import excutils
from taskflow import exceptions
from taskflow.persistence import backends

from masakari.compute import nova
import masakari.conf
from masakari.engine import driver
from masakari.engine.drivers.taskflow import base
from masakari.engine.drivers.taskflow import host_failure
from masakari.engine.drivers.taskflow import instance_failure
from masakari.engine.drivers.taskflow import process_failure
from masakari import exception
from masakari.i18n import _
from masakari import objects
from masakari.objects import fields


CONF = masakari.conf.CONF
TASKFLOW_CONF = CONF.taskflow_driver_recovery_flows
PERSISTENCE_BACKEND = CONF.taskflow.connection
LOG = logging.getLogger(__name__)


class TaskFlowDriver(driver.NotificationDriver):
    def __init__(self):
        super(TaskFlowDriver, self).__init__()

    def _execute_auto_workflow(self, context, novaclient, process_what):
        flow_engine = host_failure.get_auto_flow(context, novaclient,
                                                 process_what)

        # Attaching this listener will capture all of the notifications
        # that taskflow sends out and redirect them to a more useful
        # log for masakari's debugging (or error reporting) usage.
        with base.DynamicLogListener(flow_engine, logger=LOG):
            flow_engine.run()

    def _execute_rh_workflow(self, context, novaclient, process_what,
                             **kwargs):
        if not kwargs['reserved_host_list']:
            msg = _('No reserved_hosts available for evacuation.')
            raise exception.ReservedHostsUnavailable(message=msg)

        process_what['reserved_host_list'] = kwargs.pop('reserved_host_list')
        flow_engine = host_failure.get_rh_flow(context, novaclient,
                                               process_what,
                                               **kwargs)

        with base.DynamicLogListener(flow_engine, logger=LOG):
            try:
                flow_engine.run()
            except exception.LockAlreadyAcquired as ex:
                raise exception.HostRecoveryFailureException(ex.message)

    def _execute_auto_priority_workflow(self, context, novaclient,
                                        process_what, **kwargs):
        try:
            self._execute_auto_workflow(context, novaclient, process_what)
        except Exception as ex:
            with excutils.save_and_reraise_exception(reraise=False) as ctxt:
                if isinstance(ex, exception.SkipHostRecoveryException):
                    ctxt.reraise = True
                    return

                # Caught generic Exception to make sure that any failure
                # should lead to execute 'reserved_host' recovery workflow.
                msg = ("Failed to evacuate all instances from "
                       "failed_host: '%(failed_host)s' using "
                       "'%(auto)s' workflow, retrying using "
                       "'%(reserved_host)s' workflow.")
                LOG.warning(msg, {
                    'failed_host': process_what['host_name'],
                    'auto': fields.FailoverSegmentRecoveryMethod.AUTO,
                    'reserved_host':
                        fields.FailoverSegmentRecoveryMethod.RESERVED_HOST
                })
                self._execute_rh_workflow(context,
                                          novaclient, process_what,
                                          **kwargs)

    def _execute_rh_priority_workflow(self, context, novaclient, process_what,
                                      **kwargs):
        try:
            self._execute_rh_workflow(context, novaclient, process_what,
                                      **kwargs)
        except Exception as ex:
            with excutils.save_and_reraise_exception(reraise=False) as ctxt:
                if isinstance(ex, exception.SkipHostRecoveryException):
                    ctxt.reraise = True
                    return

                # Caught generic Exception to make sure that any failure
                # should lead to execute 'auto' recovery workflow.
                msg = ("Failed to evacuate all instances from "
                       "failed_host '%(failed_host)s' using "
                       "'%(reserved_host)s' workflow, retrying using "
                       "'%(auto)s' workflow")
                LOG.warning(msg, {
                    'failed_host': process_what['host_name'],
                    'reserved_host':
                        fields.FailoverSegmentRecoveryMethod.RESERVED_HOST,
                    'auto': fields.FailoverSegmentRecoveryMethod.AUTO
                })
                self._execute_auto_workflow(context, novaclient, process_what)

    def execute_host_failure(self, context, host_name, recovery_method,
                             notification_uuid, **kwargs):
        novaclient = nova.API()
        # get flow for host failure
        process_what = {
            'host_name': host_name,
            'notification_uuid': notification_uuid
        }

        try:
            if recovery_method == fields.FailoverSegmentRecoveryMethod.AUTO:
                self._execute_auto_workflow(context, novaclient, process_what)
            elif recovery_method == (
                    fields.FailoverSegmentRecoveryMethod.RESERVED_HOST):
                self._execute_rh_workflow(context, novaclient, process_what,
                                          **kwargs)
            elif recovery_method == (
                    fields.FailoverSegmentRecoveryMethod.AUTO_PRIORITY):
                self._execute_auto_priority_workflow(
                    context, novaclient,
                    process_what, **kwargs)
            else:
                self._execute_rh_priority_workflow(context, novaclient,
                                                   process_what, **kwargs)
        except Exception as exc:
            with excutils.save_and_reraise_exception(reraise=False) as ctxt:
                if isinstance(exc, (exception.SkipHostRecoveryException,
                                    exception.HostRecoveryFailureException,
                                    exception.ReservedHostsUnavailable)):
                    ctxt.reraise = True
                    return
                msg = _("Failed to execute host failure flow for "
                        "notification '%s'.") % notification_uuid
                raise exception.MasakariException(msg)

    def execute_instance_failure(self, context, instance_uuid,
                                 notification_uuid):
        novaclient = nova.API()
        # get flow for instance failure
        process_what = {
            'instance_uuid': instance_uuid,
            'notification_uuid': notification_uuid
        }

        try:
            flow_engine = instance_failure.get_instance_recovery_flow(
                context, novaclient, process_what)
        except Exception:
            msg = _("Failed to create instance failure flow for "
                    "notification '%s'.") % notification_uuid
            LOG.exception(msg)
            raise exception.MasakariException(msg)

        # Attaching this listener will capture all of the notifications that
        # taskflow sends out and redirect them to a more useful log for
        # masakari's debugging (or error reporting) usage.
        with base.DynamicLogListener(flow_engine, logger=LOG):
            try:
                flow_engine.run()
            except Exception as exc:
                with excutils.save_and_reraise_exception(reraise=False) as e:
                    if isinstance(
                            exc, (exception.SkipInstanceRecoveryException,
                                  exception.IgnoreInstanceRecoveryException,
                                  exception.InstanceRecoveryFailureException)):
                        e.reraise = True
                        return
                    msg = _("Failed to execute instance failure flow for "
                            "notification '%s'.") % notification_uuid
                    raise exception.MasakariException(msg)

    def execute_process_failure(self, context, process_name, host_name,
                                notification_uuid):
        novaclient = nova.API()
        # get flow for process failure
        process_what = {
            'process_name': process_name,
            'host_name': host_name,
            'notification_uuid': notification_uuid
        }

        # TODO(abhishekk) We need to create a map for process_name and
        # respective python-client so that we can pass appropriate client
        # as a input to the process.
        if process_name == "nova-compute":
            recovery_flow = process_failure.get_compute_process_recovery_flow
        else:
            LOG.warning("Skipping recovery for process: %s.",
                        process_name)
            raise exception.SkipProcessRecoveryException()

        try:
            flow_engine = recovery_flow(context, novaclient, process_what)
        except Exception:
            msg = _("Failed to create process failure flow for "
                    "notification '%s'.") % notification_uuid
            LOG.exception(msg)
            raise exception.MasakariException(msg)

        # Attaching this listener will capture all of the notifications that
        # taskflow sends out and redirect them to a more useful log for
        # masakari's debugging (or error reporting) usage.
        with base.DynamicLogListener(flow_engine, logger=LOG):
            try:
                flow_engine.run()
            except Exception as exc:
                with excutils.save_and_reraise_exception(reraise=False) as e:
                    if isinstance(
                            exc, exception.ProcessRecoveryFailureException):
                        e.reraise = True
                        return
                    msg = _("Failed to execute instance failure flow for "
                            "notification '%s'.") % notification_uuid
                    raise exception.MasakariException(msg)

    @contextlib.contextmanager
    def upgrade_backend(self, persistence_backend):
        try:
            backend = backends.fetch(persistence_backend)
            with contextlib.closing(backend.get_connection()) as conn:
                conn.upgrade()
        except exceptions.NotFound as e:
            raise e

    def _get_taskflow_sequence(self, context, recovery_method, notification):
        # Get the taskflow sequence based on the recovery method.

        novaclient = nova.API()
        task_list = []

        # Get linear task flow based on notification type
        if notification.type == fields.NotificationType.VM:
            tasks = TASKFLOW_CONF.instance_failure_recovery_tasks
        elif notification.type == fields.NotificationType.PROCESS:
            tasks = TASKFLOW_CONF.process_failure_recovery_tasks
        elif notification.type == fields.NotificationType.COMPUTE_HOST:
            if recovery_method in [
                    fields.FailoverSegmentRecoveryMethod.AUTO,
                    fields.FailoverSegmentRecoveryMethod.AUTO_PRIORITY]:
                tasks = TASKFLOW_CONF.host_auto_failure_recovery_tasks
            elif recovery_method in [
                    fields.FailoverSegmentRecoveryMethod.RESERVED_HOST,
                    fields.FailoverSegmentRecoveryMethod.RH_PRIORITY]:
                tasks = TASKFLOW_CONF.host_rh_failure_recovery_tasks

        for plugin in base.get_recovery_flow(
                tasks['pre'], context=context, novaclient=novaclient,
                update_host_method=None):
            task_list.append(plugin.name)

        for plugin in base.get_recovery_flow(
                tasks['main'], context=context, novaclient=novaclient,
                update_host_method=None):
            task_list.append(plugin.name)

        for plugin in base.get_recovery_flow(
                tasks['post'], context=context, novaclient=novaclient,
                update_host_method=None):
            task_list.append(plugin.name)

        return task_list

    def get_notification_recovery_workflow_details(self, context,
                                                   recovery_method,
                                                   notification):
        """Retrieve progress details in notification"""

        backend = backends.fetch(PERSISTENCE_BACKEND)
        with contextlib.closing(backend.get_connection()) as conn:
            progress_details = []
            flow_details = conn.get_flows_for_book(
                notification.notification_uuid)
            for flow in flow_details:
                od = OrderedDict()
                atom_details = list(conn.get_atoms_for_flow(flow.uuid))

                # TODO(ShilpaSD): In case recovery_method is auto_priority/
                # rh_priority, there is no way to figure out whether the
                # recovery was done successfully using AUTO or RH flow.
                # Taskflow stores 'retry_instance_evacuate_engine_retry' task
                # in case of RH flow so if
                # 'retry_instance_evacuate_engine_retry' is stored in the
                # given flow details then the sorting of task details should
                # happen based on the RH flow.
                # This logic won't be required after LP #1815738 is fixed.
                if recovery_method in ['AUTO_PRIORITY', 'RH_PRIORITY']:
                    persisted_task_list = [atom.name for atom in
                                           atom_details]
                    if ('retry_instance_evacuate_engine_retry' in
                            persisted_task_list):
                        recovery_method = (
                            fields.FailoverSegmentRecoveryMethod.
                            RESERVED_HOST)
                    else:
                        recovery_method = (
                            fields.FailoverSegmentRecoveryMethod.AUTO)

                # TODO(ShilpaSD): Taskflow doesn't support to return task
                # details in the same sequence in which all tasks are
                # executed. Reported this issue in LP #1815738. To resolve
                # this issue load the tasks based on the recovery method and
                # later sort it based on this task list so progress_details
                # can be returned in the expected order.
                task_list = self._get_taskflow_sequence(context,
                                                        recovery_method,
                                                        notification)

                for task in task_list:
                    for atom in atom_details:
                        if task == atom.name:
                            od[atom.name] = atom

                for key, value in od.items():
                    # Add progress_details only if tasks are executed and meta
                    # is available in which progress_details are stored.
                    if value.meta and value.meta.get("progress_details"):
                        progress_details_obj = (
                            objects.NotificationProgressDetails.create(
                                value.name,
                                value.meta['progress'],
                                value.meta['progress_details']['details']
                                ['progress_details'],
                                value.state))

                        progress_details.append(progress_details_obj)

        return progress_details



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial revision

Revision ID: 8f848eb45d03
Revises:
Create Date: 2023-07-13 12:00:07.851502
"""

from alembic import op
from oslo_db.sqlalchemy import types as oslo_db_types
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '8f848eb45d03'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        'failover_segments',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column(
            'deleted',
            oslo_db_types.SoftDeleteInteger(),
            nullable=True,
        ),
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('service_type', sa.String(length=255), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column(
            'recovery_method',
            sa.Enum(
                'auto',
                'reserved_host',
                'auto_priority',
                'rh_priority',
                name='recovery_methods',
            ),
            nullable=False,
        ),
        sa.Column('enabled', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'name', 'deleted', name='uniq_segment0name0deleted'
        ),
        sa.UniqueConstraint('uuid', name='uniq_segments0uuid'),
    )
    op.create_index(
        'segments_service_type_idx',
        'failover_segments',
        ['service_type'],
        unique=False,
    )
    op.create_table(
        'notifications',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column(
            'deleted',
            oslo_db_types.SoftDeleteInteger(),
            nullable=True,
        ),
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('notification_uuid', sa.String(length=36), nullable=False),
        sa.Column('generated_time', sa.DateTime(), nullable=False),
        sa.Column('source_host_uuid', sa.String(length=36), nullable=False),
        sa.Column('type', sa.String(length=36), nullable=False),
        sa.Column('payload', sa.Text(), nullable=True),
        sa.Column(
            'status',
            sa.Enum(
                'new',
                'running',
                'error',
                'failed',
                'ignored',
                'finished',
                name='notification_status',
            ),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'notification_uuid', name='uniq_notification0uuid'
        ),
    )
    op.create_table(
        'hosts',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column(
            'deleted',
            oslo_db_types.SoftDeleteInteger(),
            nullable=True,
        ),
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('reserved', sa.Boolean(), nullable=True),
        sa.Column('type', sa.String(length=255), nullable=False),
        sa.Column('control_attributes', sa.Text(), nullable=False),
        sa.Column('failover_segment_id', sa.String(length=36), nullable=False),
        sa.Column('on_maintenance', sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ['failover_segment_id'],
            ['failover_segments.uuid'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'deleted', name='uniq_host0name0deleted'),
        sa.UniqueConstraint('uuid', name='uniq_host0uuid'),
    )
    op.create_index('hosts_type_idx', 'hosts', ['type'], unique=False)



# Copyright (c) 2018 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from masakari.notifications.objects import base
from masakari.objects import base as masakari_base
from masakari.objects import fields


@masakari_base.MasakariObjectRegistry.register_notification
class ExceptionPayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'module_name': fields.StringField(),
        'function_name': fields.StringField(),
        'exception': fields.StringField(),
        'exception_message': fields.StringField(),
        'traceback': fields.StringField()
    }

    @classmethod
    def from_exc_and_traceback(cls, fault, traceback):
        trace = inspect.trace()
        # FIXME(mgoddard): In some code paths we reach this point without being
        # inside an exception handler. This results in inspect.trace()
        # returning an empty list. Ideally we should only end up here from an
        # exception handler.
        if trace:
            trace = trace[-1]
            # TODO(gibi): apply strutils.mask_password on exception_message and
            # consider emitting the exception_message only if the safe flag is
            # true in the exception like in the REST API
            module = inspect.getmodule(trace[0])
            function_name = trace[3]
        else:
            module = None
            function_name = 'unknown'
        module_name = module.__name__ if module else 'unknown'
        return cls(
            function_name=function_name,
            module_name=module_name,
            exception=fault.__class__.__name__,
            exception_message=str(fault),
            traceback=traceback)


@base.notification_sample('error-exception.json')
@masakari_base.MasakariObjectRegistry.register_notification
class ExceptionNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'payload': fields.ObjectField('ExceptionPayload')
    }



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields


# Import fields from oslo.versionedobjects
BooleanField = fields.BooleanField
IntegerField = fields.IntegerField
StringField = fields.StringField
EnumField = fields.EnumField
UUIDField = fields.UUIDField
DateTimeField = fields.DateTimeField
DictOfStringsField = fields.DictOfStringsField
ObjectField = fields.ObjectField
BaseEnumField = fields.BaseEnumField
ListOfObjectsField = fields.ListOfObjectsField
ListOfStringsField = fields.ListOfStringsField
FloatField = fields.FloatField
ListOfDictOfNullableStringsField = fields.ListOfDictOfNullableStringsField


Field = fields.Field
Enum = fields.Enum
FieldType = fields.FieldType


class BaseMasakariEnum(Enum):
    def __init__(self, **kwargs):
        super(BaseMasakariEnum, self).__init__(
            valid_values=self.__class__.ALL)


class FailoverSegmentRecoveryMethod(Enum):
    """Represents possible recovery_methods for failover segment."""

    AUTO = "auto"
    RESERVED_HOST = "reserved_host"
    AUTO_PRIORITY = "auto_priority"
    RH_PRIORITY = "rh_priority"

    ALL = (AUTO, RESERVED_HOST, AUTO_PRIORITY, RH_PRIORITY)

    def __init__(self):
        super(FailoverSegmentRecoveryMethod,
              self).__init__(valid_values=FailoverSegmentRecoveryMethod.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class NotificationType(Enum):
    """Represents possible notification types."""

    COMPUTE_HOST = "COMPUTE_HOST"
    VM = "VM"
    PROCESS = "PROCESS"

    ALL = (COMPUTE_HOST, VM, PROCESS)

    def __init__(self):
        super(NotificationType,
              self).__init__(valid_values=NotificationType.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class EventType(Enum):
    """Represents possible event types."""

    STARTED = "STARTED"
    STOPPED = "STOPPED"

    ALL = (STARTED, STOPPED)

    def __init__(self):
        super(EventType,
              self).__init__(valid_values=EventType.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class HostStatusType(Enum):
    """Represents possible event types for Host status."""

    NORMAL = "NORMAL"
    UNKNOWN = "UNKNOWN"

    ALL = (NORMAL, UNKNOWN)

    def __init__(self):
        super(HostStatusType,
              self).__init__(valid_values=HostStatusType.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class ClusterStatusType(Enum):
    """Represents possible event types for Cluster status."""

    ONLINE = "ONLINE"
    OFFLINE = "OFFLINE"

    ALL = (ONLINE, OFFLINE)

    def __init__(self):
        super(ClusterStatusType,
              self).__init__(valid_values=ClusterStatusType.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class NotificationStatus(Enum):
    """Represents possible statuses for notifications."""

    NEW = "new"
    RUNNING = "running"
    ERROR = "error"
    FAILED = "failed"
    IGNORED = "ignored"
    FINISHED = "finished"

    ALL = (NEW, RUNNING, ERROR, FAILED, IGNORED, FINISHED)

    def __init__(self):
        super(NotificationStatus,
              self).__init__(valid_values=NotificationStatus.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class EventNotificationAction(Enum):
    # Actions of segments
    SEGMENT_CREATE = 'segment.create'
    SEGMENT_UPDATE = 'segment.update'
    SEGMENT_DELETE = 'segment.delete'

    # Actions of hosts
    HOST_CREATE = 'host.create'
    HOST_UPDATE = 'host.update'
    HOST_DELETE = 'host.delete'

    # Actions of notifications
    NOTIFICATION_CREATE = 'notification.create'
    NOTIFICATION_PROCESS = 'notification.process'

    ALL = (SEGMENT_CREATE, SEGMENT_UPDATE, SEGMENT_DELETE, HOST_CREATE,
           HOST_UPDATE, HOST_DELETE, NOTIFICATION_CREATE,
           NOTIFICATION_PROCESS)

    def __init__(self):
        super(EventNotificationAction,
              self).__init__(valid_values=EventNotificationAction.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class EventNotificationPriority(BaseMasakariEnum):
    AUDIT = 'audit'
    CRITICAL = 'critical'
    DEBUG = 'debug'
    INFO = 'info'
    ERROR = 'error'
    SAMPLE = 'sample'
    WARN = 'warn'

    ALL = (AUDIT, CRITICAL, DEBUG, INFO, ERROR, SAMPLE, WARN)


class EventNotificationPhase(Enum):
    START = 'start'
    END = 'end'
    ERROR = 'error'

    ALL = (START, END, ERROR)

    def __init__(self):
        super(EventNotificationPhase,
              self).__init__(valid_values=EventNotificationPhase.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class FailoverSegmentRecoveryMethodField(BaseEnumField):
    AUTO_TYPE = FailoverSegmentRecoveryMethod()


class NotificationTypeField(BaseEnumField):
    AUTO_TYPE = NotificationType()


class NotificationStatusField(BaseEnumField):
    AUTO_TYPE = NotificationStatus()


class EventNotificationActionField(BaseEnumField):
    AUTO_TYPE = EventNotificationAction()


class EventNotificationPriorityField(BaseEnumField):
    AUTO_TYPE = EventNotificationPriority()


class EventNotificationPhaseField(BaseEnumField):
    AUTO_TYPE = EventNotificationPhase()


class VMoveType(Enum):
    """Represents possible types for VMoves."""

    EVACUATION = "evacuation"
    MIGRATION = "migration"
    LIVE_MIGRATION = "live_migration"

    ALL = (EVACUATION, MIGRATION, LIVE_MIGRATION)

    def __init__(self):
        super(VMoveType,
              self).__init__(valid_values=VMoveType.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class VMoveTypeField(BaseEnumField):
    AUTO_TYPE = VMoveType()


class VMoveStatus(Enum):
    """Represents possible statuses for VMoves."""

    PENDING = "pending"
    ONGOING = "ongoing"
    IGNORED = "ignored"
    FAILED = "failed"
    SUCCEEDED = "succeeded"

    ALL = (PENDING, ONGOING, IGNORED, FAILED, SUCCEEDED)

    def __init__(self):
        super(VMoveStatus,
              self).__init__(valid_values=VMoveStatus.ALL)

    @classmethod
    def index(cls, value):
        """Return an index into the Enum given a value."""
        return cls.ALL.index(value)

    @classmethod
    def from_index(cls, index):
        """Return the Enum value at a given index."""
        return cls.ALL[index]


class VMoveStatusField(BaseEnumField):
    AUTO_TYPE = VMoveStatus()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add vm moves table

Revision ID: 8bdf5929c5a6
Revises: 8f848eb45d03
Create Date: 2023-07-13 12:13:42.240598
"""

from alembic import op
from oslo_db.sqlalchemy import types as oslo_db_types
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '8bdf5929c5a6'
down_revision = '8f848eb45d03'
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        'vmoves',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column(
            'deleted',
            oslo_db_types.SoftDeleteInteger(),
            nullable=True,
        ),
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('notification_uuid', sa.String(length=36), nullable=False),
        sa.Column('instance_uuid', sa.String(length=36), nullable=False),
        sa.Column('instance_name', sa.String(length=255), nullable=False),
        sa.Column('source_host', sa.String(length=255), nullable=True),
        sa.Column('dest_host', sa.String(length=255), nullable=True),
        sa.Column('start_time', sa.DateTime(), nullable=True),
        sa.Column('end_time', sa.DateTime(), nullable=True),
        sa.Column('type', sa.String(length=36), nullable=True),
        sa.Column('status', sa.String(length=255), nullable=True),
        sa.Column('message', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_vmove0uuid'),
    )



# Copyright (c) 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The Host API extension."""

from http import HTTPStatus

from oslo_utils import encodeutils
from oslo_utils import strutils
from webob import exc

from masakari.api.openstack import common
from masakari.api.openstack import extensions
from masakari.api.openstack.ha.schemas import hosts as schema
from masakari.api.openstack.ha.views import hosts as views_hosts
from masakari.api.openstack import wsgi
from masakari.api import validation
from masakari import exception
from masakari.ha import api as host_api
from masakari.i18n import _
from masakari import objects
from masakari.policies import hosts as host_policies

ALIAS = "os-hosts"


class HostsController(wsgi.Controller):
    """The Host API controller for the OpenStack API."""

    def __init__(self):
        self.api = host_api.HostAPI()

    @extensions.expected_errors((HTTPStatus.BAD_REQUEST, HTTPStatus.FORBIDDEN,
                                 HTTPStatus.NOT_FOUND))
    def index(self, req, segment_id):
        """Returns a list a hosts."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'index')

        try:
            filters = {}
            limit, marker = common.get_limit_and_marker(req)
            sort_keys, sort_dirs = common.get_sort_params(req.params)

            segment = objects.FailoverSegment.get_by_uuid(context,
                                                          segment_id)

            filters['failover_segment_id'] = segment.uuid
            if 'name' in req.params:
                filters['name'] = req.params['name']

            if 'type' in req.params:
                filters['type'] = req.params['type']

            if 'control_attributes' in req.params:
                filters['control_attributes'] = req.params[
                    'control_attributes']

            if 'on_maintenance' in req.params:
                try:
                    filters['on_maintenance'] = strutils.bool_from_string(
                        req.params['on_maintenance'], strict=True)
                except ValueError as ex:
                    msg = _("Invalid value for on_maintenance: "
                            "%s") % encodeutils.exception_to_unicode(ex)
                    raise exc.HTTPBadRequest(explanation=msg)

            if 'reserved' in req.params:
                try:
                    filters['reserved'] = strutils.bool_from_string(
                        req.params['reserved'], strict=True)
                except ValueError as ex:
                    msg = _("Invalid value for reserved: "
                            "%s") % encodeutils.exception_to_unicode(ex)
                    raise exc.HTTPBadRequest(explanation=msg)

            hosts = self.api.get_all(context, filters=filters,
                                     sort_keys=sort_keys, sort_dirs=sort_dirs,
                                     limit=limit, marker=marker)
        except exception.MarkerNotFound as ex:
            raise exc.HTTPBadRequest(explanation=ex.format_message())
        except exception.Invalid as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except exception.FailoverSegmentNotFound as ex:
            raise exc.HTTPNotFound(explanation=ex.format_message())

        builder = views_hosts.get_view_builder(req)
        return builder.build_hosts(hosts)

    @wsgi.response(HTTPStatus.CREATED)
    @extensions.expected_errors((HTTPStatus.BAD_REQUEST, HTTPStatus.FORBIDDEN,
                                 HTTPStatus.NOT_FOUND, HTTPStatus.CONFLICT))
    @validation.schema(schema.create)
    def create(self, req, segment_id, body):
        """Creates a host."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'create')
        host_data = body.get('host')
        try:
            host = self.api.create_host(context, segment_id, host_data)
        except exception.ComputeNotFoundByName as e:
            raise exc.HTTPBadRequest(explanation=e.message)
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.HostExists as e:
            raise exc.HTTPConflict(explanation=e.format_message())

        builder = views_hosts.get_view_builder(req)
        return {'host': builder.build_host(host)}

    @extensions.expected_errors((HTTPStatus.FORBIDDEN, HTTPStatus.NOT_FOUND))
    def show(self, req, segment_id, id):
        """Shows the details of a host."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'detail')
        try:
            host = self.api.get_host(context, segment_id, id)
        except exception.HostNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())

        builder = views_hosts.get_view_builder(req)
        return {'host': builder.build_host(host)}

    @extensions.expected_errors((HTTPStatus.BAD_REQUEST, HTTPStatus.FORBIDDEN,
                                 HTTPStatus.NOT_FOUND, HTTPStatus.CONFLICT))
    @validation.schema(schema.update)
    def update(self, req, segment_id, id, body):
        """Updates the existing host."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'update')
        host_data = body.get('host')
        try:
            host = self.api.update_host(context, segment_id, id, host_data)
        except exception.ComputeNotFoundByName as e:
            raise exc.HTTPBadRequest(explanation=e.message)
        except exception.HostNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except (exception.HostExists, exception.Conflict) as e:
            raise exc.HTTPConflict(explanation=e.format_message())

        builder = views_hosts.get_view_builder(req)
        return {'host': builder.build_host(host)}

    @wsgi.response(HTTPStatus.NO_CONTENT)
    @extensions.expected_errors((HTTPStatus.FORBIDDEN, HTTPStatus.NOT_FOUND,
                                 HTTPStatus.CONFLICT))
    def delete(self, req, segment_id, id):
        """Removes a host by id."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'delete')
        try:
            self.api.delete_host(context, segment_id, id)
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.HostNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.Conflict as e:
            raise exc.HTTPConflict(explanation=e.format_message())


class Hosts(extensions.V1APIExtensionBase):
    """Hosts controller"""

    name = "Hosts"
    alias = ALIAS
    version = 1

    def get_resources(self):
        parent = {'member_name': 'segment',
                  'collection_name': 'segments'}
        resources = [
            extensions.ResourceExtension(
                'hosts', HostsController(), parent=parent,
                member_name='host')]

        return resources

    def get_controller_extensions(self):
        return []



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Generic Node base class for all workers that run on hosts."""

import os
import random
import sys

from oslo_concurrency import processutils
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import service
from oslo_utils import importutils

import masakari.conf
from masakari import context
from masakari import coordination as masakari_coordination
from masakari import exception
from masakari.i18n import _
from masakari.objects import base as objects_base
from masakari import rpc
from masakari import utils
from masakari import version
from masakari import wsgi


LOG = logging.getLogger(__name__)

CONF = masakari.conf.CONF


class Service(service.Service):
    """Service object for binaries running on hosts.

    A service takes a manager and enables rpc by listening to queues based
    on topic. It also periodically runs tasks on the manager.
    """

    def __init__(self, host, binary, topic, manager,
                 periodic_enable=None, periodic_fuzzy_delay=None,
                 periodic_interval_max=None):
        super(Service, self).__init__()

        if not rpc.initialized():
            rpc.init(CONF)

        self.host = host
        self.binary = binary
        self.topic = topic
        self.manager_class_name = manager
        manager_class = importutils.import_class(self.manager_class_name)
        self.rpcserver = None
        self.manager = manager_class(host=self.host)
        self.periodic_enable = periodic_enable
        self.periodic_fuzzy_delay = periodic_fuzzy_delay
        self.periodic_interval_max = periodic_interval_max

    def __repr__(self):
        return "<%(cls_name)s: host=%(host)s, binary=%(binary)s, " \
               "manager_class_name=%(manager)s>" %\
               {
                   'cls_name': self.__class__.__name__,
                   'host': self.host,
                   'binary': self.binary,
                   'manager': self.manager_class_name
               }

    def start(self):
        verstr = version.version_string_with_package()
        LOG.info('Starting %(topic)s (version %(version)s)', {
            'topic': self.topic,
            'version': verstr
        })
        self.basic_config_check()

        LOG.debug("Creating RPC server for service %s", self.topic)

        target = messaging.Target(topic=self.topic, server=self.host)
        endpoints = [self.manager]
        serializer = objects_base.MasakariObjectSerializer()
        self.rpcserver = rpc.get_server(target, endpoints, serializer)
        self.rpcserver.start()

        if self.periodic_enable:
            if self.periodic_fuzzy_delay:
                initial_delay = random.randint(0, self.periodic_fuzzy_delay)
            else:
                initial_delay = None

            self.tg.add_dynamic_timer(
                self.periodic_tasks,
                initial_delay=initial_delay,
                periodic_interval_max=self.periodic_interval_max)

    def __getattr__(self, key):
        manager = self.__dict__.get('manager', None)
        return getattr(manager, key)

    @classmethod
    def create(cls, host=None, binary=None, topic=None, manager=None,
               periodic_enable=None, periodic_fuzzy_delay=None,
               periodic_interval_max=None):
        """Instantiates class and passes back application object.

        :param host: defaults to CONF.host
        :param binary: defaults to basename of executable
        :param topic: defaults to bin_name - 'masakari-' part
        :param manager: defaults to CONF.<topic>_manager
        :param periodic_enable: defaults to CONF.periodic_enable
        :param periodic_fuzzy_delay: defaults to CONF.periodic_fuzzy_delay
        :param periodic_interval_max: if set, the max time to wait between runs

        """
        if not host:
            host = CONF.host
        if not binary:
            binary = os.path.basename(sys.argv[0])
        if not topic:
            topic = binary.rpartition('masakari-')[2]
        if not manager:
            manager_cls = ('%s_manager' %
                           binary.rpartition('masakari-')[2])
            manager = CONF.get(manager_cls, None)
        if periodic_enable is None:
            periodic_enable = CONF.periodic_enable
        if periodic_fuzzy_delay is None:
            periodic_fuzzy_delay = CONF.periodic_fuzzy_delay
        if periodic_interval_max is None:
            periodic_interval_max = CONF.periodic_interval_max

        service_obj = cls(host, binary, topic, manager,
                          periodic_enable=periodic_enable,
                          periodic_fuzzy_delay=periodic_fuzzy_delay,
                          periodic_interval_max=periodic_interval_max)

        return service_obj

    def kill(self):
        """Destroy the service object in the datastore.

        NOTE: Although this method is not used anywhere else than tests, it is
        convenient to have it here, so the tests might easily and in clean way
        stop and remove the service_ref.

        """
        self.stop()

    def stop(self):
        # Try to shut the connection down, but if we get any sort of
        # errors, go ahead and ignore them.. as we're shutting down anyway
        try:
            self.rpcserver.stop()
        except Exception:
            pass
        super(Service, self).stop()

    def periodic_tasks(self, raise_on_error=False):
        """Tasks to be run at a periodic interval."""
        ctxt = context.get_admin_context()
        return self.manager.periodic_tasks(ctxt, raise_on_error=raise_on_error)

    def basic_config_check(self):
        """Perform basic config checks before starting processing."""
        # Make sure the tempdir exists and is writable
        try:
            with utils.tempdir():
                pass
        except Exception as e:
            LOG.error('Temporary directory is invalid: %s', e)
            sys.exit(1)

    def reset(self):
        self.manager.reset()


class WSGIService(service.Service):
    """Provides ability to launch API from a 'paste' configuration."""

    def __init__(self, name, loader=None, use_ssl=False, max_url_len=None,
                 coordination=False):
        """Initialize, but do not start the WSGI server.

        :param name: The name of the WSGI server given to the loader.
        :param loader: Loads the WSGI application using the given name.
        :returns: None
        """
        self.name = name
        self.binary = 'masakari-%s' % name
        self.topic = None
        self.loader = loader or wsgi.Loader()
        self.app = self.loader.load_app(name)
        self.host = getattr(CONF, '%s_listen' % name, "0.0.0.0")
        self.port = getattr(CONF, '%s_listen_port' % name, 0)

        self.workers = (getattr(CONF, '%s_workers' % name, None) or
                        processutils.get_worker_count())

        if self.workers and self.workers < 1:
            worker_name = '%s_workers' % name
            msg = (_("%(worker_name)s value of %(workers)s is invalid, "
                     "must be greater than 0") %
                   {'worker_name': worker_name,
                    'workers': str(self.workers)})
            raise exception.InvalidInput(msg)

        self.use_ssl = use_ssl
        self.server = wsgi.Server(name,
                                  self.app,
                                  host=self.host,
                                  port=self.port,
                                  use_ssl=self.use_ssl,
                                  max_url_len=max_url_len)
        self.coordination = coordination

    def reset(self):
        """Reset server greenpool size to default.

        :returns: None

        """
        self.server.reset()

    def start(self):
        """Start serving this service using loaded configuration.

        Also, retrieve updated port number in case '0' was passed in, which
        indicates a random port should be used.

        :returns: None

        """
        if self.coordination:
            masakari_coordination.COORDINATOR.start()
        self.server.start()

    def stop(self):
        """Stop serving this API.

        :returns: None

        """
        if self.coordination:
            try:
                masakari_coordination.COORDINATOR.stop()
            except Exception as error:
                LOG.warning('Error occurred during masakari coordination was '
                            'stopped: %s', error)
        self.server.stop()

    def wait(self):
        """Wait for the service to stop serving this API.

        :returns: None

        """
        self.server.wait()


def process_launcher():
    return service.ProcessLauncher(CONF, restart_method='mutate')


# NOTE: the global launcher is to maintain the existing
#       functionality of calling service.serve +
#       service.wait
_launcher = None


def serve(server, workers=None):
    global _launcher
    if _launcher:
        raise RuntimeError(_('serve() can only be called once'))

    _launcher = service.launch(CONF, server, workers=workers,
                               restart_method='mutate')


def wait():
    _launcher.wait()



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
import routes
import webob.dec

from masakari.api import api_version_request as api_version
from masakari.api import auth as api_auth
from masakari.api import openstack as openstack_api
from masakari.api.openstack import ha
from masakari.api.openstack.ha import versions
from masakari.api.openstack import wsgi as os_wsgi
from masakari.api import urlmap
from masakari import context
from masakari.tests import uuidsentinel
from masakari import wsgi


@webob.dec.wsgify
def fake_wsgi(self, req):
    return self.application


def wsgi_app_v1(fake_auth_context=None, init_only=None):

    inner_app_v1 = ha.APIRouterV1()

    if fake_auth_context is not None:
        ctxt = fake_auth_context
    else:
        ctxt = context.RequestContext('fake', 'fake', auth_token=True)
    api_v1 = (
        openstack_api.FaultWrapper(api_auth.InjectContext(ctxt, inner_app_v1)))
    mapper = urlmap.URLMap()
    mapper['/v1'] = api_v1
    mapper['/'] = openstack_api.FaultWrapper(versions.Versions())
    return mapper


class FakeToken(object):
    id_count = 0

    def __getitem__(self, key):
        return getattr(self, key)

    def __init__(self, **kwargs):
        FakeToken.id_count += 1
        self.id = FakeToken.id_count
        for k, v in kwargs.items():
            setattr(self, k, v)


class FakeRequestContext(context.RequestContext):
    def __init__(self, *args, **kwargs):
        kwargs['auth_token'] = kwargs.get('auth_token', 'fake_auth_token')
        super(FakeRequestContext, self).__init__(*args, **kwargs)


class HTTPRequest(os_wsgi.Request):

    @staticmethod
    def blank(*args, **kwargs):
        kwargs['base_url'] = 'http://localhost/v1'
        use_admin_context = kwargs.pop('use_admin_context', False)
        project_id = kwargs.pop('project_id', uuidsentinel.fake_project_id)
        version = kwargs.pop('version', api_version.DEFAULT_API_VERSION)
        out = os_wsgi.Request.blank(*args, **kwargs)
        out.environ['masakari.context'] = FakeRequestContext(
            user_id=uuidsentinel.fake_user_id,
            project_id=project_id,
            is_admin=use_admin_context)
        out.api_version_request = api_version.APIVersionRequest(version)
        return out


class TestRouter(wsgi.Router):
    def __init__(self, controller, mapper=None):
        if not mapper:
            mapper = routes.Mapper()
        mapper.resource("test", "tests",
                        controller=os_wsgi.Resource(controller))
        super(TestRouter, self).__init__(mapper)


class FakeAuthDatabase(object):
    data = {}

    @staticmethod
    def auth_token_get(context, token_hash):
        return FakeAuthDatabase.data.get(token_hash, None)

    @staticmethod
    def auth_token_create(context, token):
        fake_token = FakeToken(created_at=timeutils.utcnow(), **token)
        FakeAuthDatabase.data[fake_token.token_hash] = fake_token
        FakeAuthDatabase.data['id_%i' % fake_token.id] = fake_token
        return fake_token

    @staticmethod
    def auth_token_destroy(context, token_id):
        token = FakeAuthDatabase.data.get('id_%i' % token_id)
        if token and token.token_hash in FakeAuthDatabase.data:
            del FakeAuthDatabase.data[token.token_hash]
            del FakeAuthDatabase.data['id_%i' % token_id]


def fake_get_available_languages():
    existing_translations = ['en_GB', 'en_AU', 'de', 'zh_CN', 'en_US']
    return existing_translations


def fake_not_implemented(*args, **kwargs):
    raise NotImplementedError()



# Copyright 2016 NTT DATA
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for process failure TaskFlow
"""

from unittest import mock

from masakari.compute import nova
from masakari import conf
from masakari import context
from masakari.engine.drivers.taskflow import process_failure
from masakari import exception
from masakari import test
from masakari.tests.unit import fakes

CONF = conf.CONF


class ProcessFailureTestCase(test.TestCase):

    def setUp(self):
        super(ProcessFailureTestCase, self).setUp()
        self.ctxt = context.get_admin_context()
        self.process_name = "nova-compute"
        self.service_host = "fake-host"
        self.novaclient = nova.API()
        self.fake_client = fakes.FakeNovaClient()
        # overriding 'wait_period_after_service_update' to 2 seconds
        # to reduce the wait period.
        self.override_config('wait_period_after_service_update', 2)
        self.disabled_reason = CONF.process_failure.service_disable_reason

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_compute_process_failure_flow(self, _mock_notify,
                                          _mock_novaclient):
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.services.create("1", host=self.service_host,
                                         binary="nova-compute",
                                         status="enabled")

        # test DisableComputeNodeTask
        task = process_failure.DisableComputeNodeTask(self.ctxt,
                                                      self.novaclient)
        task.execute(self.process_name, self.service_host)

        # test ConfirmComputeNodeDisabledTask
        task = process_failure.ConfirmComputeNodeDisabledTask(self.ctxt,
                                                              self.novaclient)
        task.execute(self.process_name, self.service_host)

        # verify service is disabled
        self.assertTrue(
            self.novaclient.is_service_disabled(self.ctxt,
                                                self.service_host,
                                                self.process_name))

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call("Confirming compute service is disabled on host: "
                      "'fake-host'"),
            mock.call("Confirmed compute service is disabled on host: "
                      "'fake-host'", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_compute_process_failure_flow_disabled_process(self,
                                                           _mock_notify,
                                                           _mock_novaclient):
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.services.create("1", host=self.service_host,
                                         binary="nova-compute",
                                         status="disabled")

        # test DisableComputeNodeTask
        task = process_failure.DisableComputeNodeTask(self.ctxt,
                                                      self.novaclient)
        with mock.patch.object(
            self.novaclient,
            'enable_disable_service') as mock_enable_disabled:
            task.execute(self.process_name, self.service_host)

        # ensure that enable_disable_service method is not called
        self.assertEqual(0, mock_enable_disabled.call_count)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call('Skipping recovery for process nova-compute as it is '
                      'already disabled', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_compute_process_failure_flow_compute_service_disabled_failed(
        self, _mock_notify, _mock_novaclient):
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.services.create("1", host=self.service_host,
                                         binary="nova-compute",
                                         status="enabled")

        def fake_is_service_disabled(context, host_name, binary):
            # assume that service is not disabled
            return False

        # test DisableComputeNodeTask
        task = process_failure.DisableComputeNodeTask(self.ctxt,
                                                      self.novaclient)
        task.execute(self.process_name, self.service_host)

        with mock.patch.object(self.novaclient, 'is_service_disabled',
                               fake_is_service_disabled):
            # test ConfirmComputeNodeDisabledTask
            task = process_failure.ConfirmComputeNodeDisabledTask(
                self.ctxt, self.novaclient)
            self.assertRaises(exception.ProcessRecoveryFailureException,
                              task.execute, self.process_name,
                              self.service_host)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call("Confirming compute service is disabled on host: "
                      "'fake-host'"),
            mock.call('Failed to disable service nova-compute', 1.0)
        ])



# Copyright 2016 NTT DATA
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock

from oslo_config import cfg
import webob.exc

from masakari.api.openstack import extensions
from masakari.api.openstack import ha
from masakari.api.openstack.ha import extension_info
from masakari import exception
from masakari import test

CONF = cfg.CONF


class fake_bad_extension(object):
    name = "fake_bad_extension"
    alias = "fake-bad"


class ExtensionLoadingTestCase(test.NoDBTestCase):

    @mock.patch('masakari.rpc.get_client')
    def test_extensions_loaded(self, mock_get_client):
        app = ha.APIRouterV1()
        self.assertIn('extensions', app._loaded_extension_info.extensions)

    def test_check_bad_extension(self):
        loaded_ext_info = extension_info.LoadedExtensionInfo()
        self.assertFalse(loaded_ext_info._check_extension(fake_bad_extension))

    @mock.patch('masakari.rpc.get_client')
    @mock.patch('masakari.api.openstack.APIRouterV1._register_resources_list')
    def test_extensions_inherit(self, mock_register, mock_get_client):
        app = ha.APIRouterV1()
        self.assertIn('extensions', app._loaded_extension_info.extensions)

        ext_no_inherits = mock_register.call_args_list[0][0][0]
        mock_register.assert_called_with(mock.ANY, mock.ANY)
        name_list = [ext.obj.alias for ext in ext_no_inherits]
        self.assertIn('extensions', name_list)

    def test_extensions_expected_error(self):
        @extensions.expected_errors(HTTPStatus.NOT_FOUND)
        def fake_func():
            raise webob.exc.HTTPNotFound()

        self.assertRaises(webob.exc.HTTPNotFound, fake_func)

    def test_extensions_expected_error_from_list(self):
        @extensions.expected_errors((HTTPStatus.NOT_FOUND,
                                     HTTPStatus.FORBIDDEN))
        def fake_func():
            raise webob.exc.HTTPNotFound()

        self.assertRaises(webob.exc.HTTPNotFound, fake_func)

    def test_extensions_unexpected_error(self):
        @extensions.expected_errors(HTTPStatus.NOT_FOUND)
        def fake_func():
            raise webob.exc.HTTPConflict()

        self.assertRaises(webob.exc.HTTPInternalServerError, fake_func)

    def test_extensions_unexpected_error_from_list(self):
        @extensions.expected_errors((HTTPStatus.NOT_FOUND,
                                     HTTPStatus.REQUEST_ENTITY_TOO_LARGE))
        def fake_func():
            raise webob.exc.HTTPConflict()

        self.assertRaises(webob.exc.HTTPInternalServerError, fake_func)

    def test_extensions_unexpected_policy_not_authorized_error(self):
        @extensions.expected_errors(HTTPStatus.NOT_FOUND)
        def fake_func():
            raise exception.PolicyNotAuthorized(action="foo")

        self.assertRaises(exception.PolicyNotAuthorized, fake_func)



# Copyright (c) 2016 NTT Data
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
:mod:`masakari` -- Cloud IaaS Platform
===================================

.. automodule:: masakari
   :platform: Unix
   :synopsis: Infrastructure-as-a-Service Cloud platform.
"""

import os

os.environ['EVENTLET_NO_GREENDNS'] = 'yes'

# NOTE(rpodolyaka): import oslo_service first, so that it makes eventlet hub
# use a monotonic clock to avoid issues with drifts of system time (see
# LP 1510234 for details)
import oslo_service  # noqa

import eventlet  # noqa
